#! /usr/bin/env python
"""SearchSource: ERIC"""
from __future__ import annotations

import typing
import urllib.parse
from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.eric.src import eric_api

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class ERICSearchSource:
    """ERIC API"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    # pylint: disable=colrev-missed-constant-usage
    source_identifier = "ID"
    search_types = [SearchType.API]
    endpoint = "colrev.eric"

    db_url = "https://eric.ed.gov/"

    ci_supported: bool = Field(default=True)
    heuristic_status = SearchSourceHeuristicStatus.oni

    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager
        self.source_operation = source_operation
        if settings:
            # ERIC as a search_source
            self.search_source = self.settings_class(**settings)
        else:
            self.search_source = colrev.settings.SearchSource(
                endpoint=self.endpoint,
                filename=Path("data/search/eric.bib"),
                search_type=SearchType.OTHER,
                search_parameters={},
                comment="",
            )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for ERIC"""

        result = {"confidence": 0.1}

        if "OWN - ERIC" in data:
            # if data.count("OWN - ERIC") > data.count("\n\n"):
            result["confidence"] = 1.0

        # Note : no features in bib file for identification

        return result

    @classmethod
    def _search_split(cls, search: str) -> str:
        if " AND " in search:
            search_parts = search.split(" AND ")
            field_values = []
            for part in search_parts:
                field, value = part.split(":")
                field = field.strip()
                value = value.strip().strip("'")
                field_value = f"{field}%3A%22{urllib.parse.quote(value)}%22"
                field_values.append(field_value)
            return " AND ".join(field_values)

        field, value = search.split(":")
        field = field.strip()
        value = value.strip().strip("'")
        field_value = f"{field}%3A%22{urllib.parse.quote(value)}%22"
        return field_value

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a)"""

        params_dict = {}
        if params:
            if params.startswith("http"):
                params_dict = {Fields.URL: params}
            else:
                for item in params.split(";"):
                    key, value = item.split("=")
                    params_dict[key] = value

        # all API searches

        if len(params_dict) == 0:
            search_source = operation.create_db_source(
                search_source_cls=cls, params=params_dict
            )

        # pylint: disable=colrev-missed-constant-usage
        elif "https://api.ies.ed.gov/eric/?" in params_dict["url"]:
            url_parsed = urllib.parse.urlparse(params_dict["url"])
            new_query = urllib.parse.parse_qs(url_parsed.query)
            search = new_query.get("search", [""])[0]
            start = new_query.get("start", ["0"])[0]
            rows = new_query.get("rows", ["2000"])[0]
            if ":" in search:
                search = ERICSearchSource._search_split(search)
            filename = operation.get_unique_filename(file_path_string=f"eric_{search}")
            search_source = colrev.settings.SearchSource(
                endpoint=cls.endpoint,
                filename=filename,
                search_type=SearchType.API,
                search_parameters={"query": search, "start": start, "rows": rows},
                comment="",
            )

        else:
            raise colrev_exceptions.PackageParameterError(
                f"Cannot add ERIC endpoint with query {params_dict}"
            )

        operation.add_source_and_search(search_source)
        return search_source

    def _run_api_search(
        self, *, eric_feed: colrev.ops.search_api_feed.SearchAPIFeed, rerun: bool
    ) -> None:

        api = eric_api.ERICAPI(params=self.search_source.search_parameters)
        for record in api.get_query_return():
            eric_feed.add_update_record(record)

        eric_feed.save()

    def search(self, rerun: bool) -> None:
        """Run a search of ERIC"""

        eric_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        if self.search_source.search_type == SearchType.API:
            self._run_api_search(eric_feed=eric_feed, rerun=rerun)
        elif self.search_source.search_type == SearchType.DB:
            self.source_operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )
        else:
            raise NotImplementedError

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def _load_nbib(self) -> dict:
        def entrytype_setter(record_dict: dict) -> None:
            if "Journal Articles" in record_dict["PT"]:
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
            else:
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC

        def field_mapper(record_dict: dict) -> None:

            key_maps = {
                ENTRYTYPES.ARTICLE: {
                    "TI": Fields.TITLE,
                    "AU": Fields.AUTHOR,
                    "DP": Fields.YEAR,
                    "JT": Fields.JOURNAL,
                    "VI": Fields.VOLUME,
                    "IP": Fields.NUMBER,
                    "PG": Fields.PAGES,
                    "AB": Fields.ABSTRACT,
                    "AID": Fields.DOI,
                    "ISSN": Fields.ISSN,
                    "OID": f"{self.endpoint}.eric_id",
                    "OT": Fields.KEYWORDS,
                    "LA": Fields.LANGUAGE,
                    "PT": "type",
                    "LID": f"{self.endpoint}.eric_url",
                }
            }

            key_map = key_maps[record_dict[Fields.ENTRYTYPE]]
            for ris_key in list(record_dict.keys()):
                if ris_key in key_map:
                    standard_key = key_map[ris_key]
                    record_dict[standard_key] = record_dict.pop(ris_key)

            if Fields.AUTHOR in record_dict and isinstance(
                record_dict[Fields.AUTHOR], list
            ):
                record_dict[Fields.AUTHOR] = " and ".join(record_dict[Fields.AUTHOR])
            if Fields.EDITOR in record_dict and isinstance(
                record_dict[Fields.EDITOR], list
            ):
                record_dict[Fields.EDITOR] = " and ".join(record_dict[Fields.EDITOR])
            if Fields.KEYWORDS in record_dict and isinstance(
                record_dict[Fields.KEYWORDS], list
            ):
                record_dict[Fields.KEYWORDS] = ", ".join(record_dict[Fields.KEYWORDS])

            record_dict.pop("type", None)
            record_dict.pop("OWN", None)
            record_dict.pop("SO", None)

            for key, value in record_dict.items():
                record_dict[key] = str(value)

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            unique_id_field="OID",
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=self.review_manager.logger,
        )

        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".nbib":
            return self._load_nbib()

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for ERIC"""

        if Fields.ISSN in record.data:
            record.data[Fields.ISSN] = record.data[Fields.ISSN].lstrip("ISSN-")

        return record



#!/usr/bin/env python3
"""Remove records, ... from CoLRev projects."""
from __future__ import annotations

from pathlib import Path

import colrev.process.operation
from colrev.constants import Fields
from colrev.constants import OperationsType
from colrev.writer.write_utils import write_file


class Remove(colrev.process.operation.Operation):
    """Remove records, ... from CoLRev projects."""

    type = OperationsType.check

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=False,
        )

    @colrev.process.operation.Operation.decorate()
    def remove_records(self, *, ids: str) -> None:
        """Remove records from CoLRev project."""

        records = self.review_manager.dataset.load_records_dict()

        for record_id in ids.split(","):
            if record_id in records:
                self.review_manager.logger.info(f" remove {record_id}")
                origins = records[record_id][Fields.ORIGIN]
                if Fields.FILE in records[record_id]:
                    print(f"manually remove file: {records[record_id]['file']}")
                del records[record_id]
                for origin in origins:
                    file, origin_id = origin.split("/")

                    search_dir = self.review_manager.paths.search
                    filepath = search_dir / Path(file)

                    origin_records = colrev.loader.load_utils.load(
                        filename=filepath,
                        logger=self.review_manager.logger,
                    )

                    if origin_id in origin_records:
                        del origin_records[origin_id]

                    write_file(records_dict=origin_records, filename=filepath)

                    self.review_manager.dataset.add_changes(filepath)

        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.dataset.create_commit(
            msg="Remove records", manual_author=False
        )



#! /usr/bin/env python
"""Functionality to determine similarity betwen records."""
from __future__ import annotations

import re
import typing

import pandas as pd
from bib_dedupe.bib_dedupe import block
from bib_dedupe.bib_dedupe import match
from bib_dedupe.bib_dedupe import prep
from rapidfuzz import fuzz

import colrev.env.utils
from colrev.constants import Fields
from colrev.constants import FieldValues

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.record.record


def get_record_change_score(
    record_a: colrev.record.record.Record, record_b: colrev.record.record.Record
) -> float:
    """Determine how much records changed"""

    # At some point, this may become more sensitive to major changes
    str_a = (
        f"{record_a.data.get(Fields.AUTHOR, '')} ({record_a.data.get(Fields.YEAR, '')}) "
        + f"{record_a.data.get(Fields.TITLE, '')}. "
        + f"{record_a.data.get(Fields.JOURNAL, '')}{record_a.data.get(Fields.BOOKTITLE, '')}, "
        + f"{record_a.data.get(Fields.VOLUME, '')} ({record_a.data.get(Fields.NUMBER, '')})"
    )
    str_b = (
        f"{record_b.data.get(Fields.AUTHOR, '')} ({record_b.data.get(Fields.YEAR, '')}) "
        + f"{record_b.data.get(Fields.TITLE, '')}. "
        + f"{record_b.data.get(Fields.JOURNAL, '')}{record_b.data.get(Fields.BOOKTITLE, '')}, "
        + f"{record_b.data.get(Fields.VOLUME, '')} ({record_b.data.get(Fields.NUMBER, '')})"
    )
    return 1 - fuzz.ratio(str_a.lower(), str_b.lower()) / 100


def container_is_abbreviated(record: colrev.record.record.Record) -> bool:
    """Check whether the container title is abbreviated"""
    if Fields.CONTAINER_TITLE in record.data:
        if record.data[Fields.CONTAINER_TITLE].count(".") > 2:
            return True
        if record.data[Fields.CONTAINER_TITLE].isupper():
            return True
    return False


def _format_authors_string_for_comparison(record: colrev.record.record.Record) -> None:
    if Fields.AUTHOR not in record.data:  # pragma: no cover
        return
    authors = str(record.data[Fields.AUTHOR]).lower()
    authors = colrev.env.utils.remove_accents(authors)

    # abbreviate first names
    # "Webster, Jane" -> "Webster, J"
    # also remove all special characters and do not include separators (and)
    authors_string = ""
    for author in authors.split(" and "):
        if "," in author:
            last_names = [
                word[0] for word in author.split(",")[1].split(" ") if len(word) > 0
            ]
            authors_string = (
                authors_string + author.split(",")[0] + " " + " ".join(last_names) + " "
            )
        else:
            authors_string = authors_string + author + " "
    authors_string = re.sub(r"[^A-Za-z0-9, ]+", "", authors_string.rstrip())
    record.data[Fields.AUTHOR] = authors_string


def _abbreviate_container_title(
    record: colrev.record.record.Record,
) -> None:
    def abbreviate_container(
        record: colrev.record.record.Record, *, min_len: int
    ) -> None:
        abbreviated_container = " ".join(
            [x[:min_len] for x in record.data[Fields.CONTAINER_TITLE].split(" ")]
        )
        record.data[Fields.CONTAINER_TITLE] = abbreviated_container

    def get_abbrev_container_min_len(record: colrev.record.record.Record) -> int:
        return min(
            len(x)
            for x in record.data[Fields.CONTAINER_TITLE].replace(".", "").split(" ")
        )

    if Fields.CONTAINER_TITLE not in record.data:
        record.data[Fields.CONTAINER_TITLE] = (
            record.data.get(Fields.JOURNAL, "")
            + record.data.get(Fields.BOOKTITLE, "")
            + record.data.get(Fields.SERIES, "")
        )

    if container_is_abbreviated(record):
        min_len = get_abbrev_container_min_len(record)
        abbreviate_container(record, min_len=min_len)


def _get_similarity_detailed(record_a: dict, record_b: dict) -> float:
    """Determine the detailed similarities between records"""
    author_similarity = (
        fuzz.ratio(record_a.get(Fields.AUTHOR, ""), record_b.get(Fields.AUTHOR, ""))
        / 100
    )

    title_similarity = (
        fuzz.ratio(
            record_a.get(Fields.TITLE, "").lower().replace(":", "").replace("-", ""),
            record_b.get(Fields.TITLE, "").lower().replace(":", "").replace("-", ""),
        )
        / 100
    )

    # partial ratio (catching 2010-10 or 2001-2002)
    year_similarity = (
        fuzz.ratio(
            str(record_a.get(Fields.YEAR, "")), str(record_b.get(Fields.YEAR, ""))
        )
        / 100
    )

    outlet_similarity = 0.0
    if record_b.get(Fields.CONTAINER_TITLE, "") and record_a.get(
        Fields.CONTAINER_TITLE, ""
    ):
        outlet_similarity = (
            fuzz.ratio(
                record_a.get(Fields.CONTAINER_TITLE, ""),
                record_b.get(Fields.CONTAINER_TITLE, ""),
            )
            / 100
        )

    if record_a.get(Fields.JOURNAL, "") not in [
        "",
        FieldValues.UNKNOWN,
    ] and record_b.get(Fields.JOURNAL, "") not in ["", FieldValues.UNKNOWN]:
        # Note: for journals papers, we expect more details
        volume_similarity = (
            1
            if (record_a.get(Fields.VOLUME, "") == record_b.get(Fields.VOLUME, ""))
            else 0
        )

        number_similarity = (
            1
            if (record_a.get(Fields.NUMBER, "") == record_b.get(Fields.NUMBER, ""))
            else 0
        )

        # Put more weight on other fields if the title is very common
        # ie., non-distinctive
        # The list is based on a large export of distinct papers, tabulated
        # according to titles and sorted by frequency
        if all(
            title
            in [
                "editorial",
                "editorial introduction",
                "editorial notes",
                "editor's comments",
                "book reviews",
                "editorial note",
                "reviewer ackowledgment",
            ]
            for title in [
                record_a.get(Fields.TITLE, "").lower(),
                record_b.get(Fields.TITLE, "").lower(),
            ]
        ):
            weights = [0.175, 0, 0.175, 0.175, 0.275, 0.2]
        else:
            weights = [0.2, 0.25, 0.13, 0.2, 0.12, 0.1]

        similarities = [
            author_similarity,
            title_similarity,
            year_similarity,
            outlet_similarity,
            volume_similarity,
            number_similarity,
        ]

    else:
        weights = [0.15, 0.75, 0.05, 0.05]
        similarities = [
            author_similarity,
            title_similarity,
            year_similarity,
            outlet_similarity,
        ]

    weighted_average = sum(
        similarities[g] * weights[g] for g in range(len(similarities))
    )
    return round(weighted_average, 4)


def _ensure_mandatory_fields(
    record_a: colrev.record.record.Record, record_b: colrev.record.record.Record
) -> None:
    mandatory_fields = [
        Fields.TITLE,
        Fields.AUTHOR,
        Fields.YEAR,
        Fields.JOURNAL,
        Fields.VOLUME,
        Fields.NUMBER,
        Fields.PAGES,
        Fields.BOOKTITLE,
    ]

    for mandatory_field in mandatory_fields:
        if (
            record_a.data.get(mandatory_field, FieldValues.UNKNOWN)
            == FieldValues.UNKNOWN
        ):
            record_a.data[mandatory_field] = ""
        if (
            record_b.data.get(mandatory_field, FieldValues.UNKNOWN)
            == FieldValues.UNKNOWN
        ):
            record_b.data[mandatory_field] = ""


def get_record_similarity(
    record_a: colrev.record.record.Record, record_b: colrev.record.record.Record
) -> float:
    """Determine the similarity between two records (their masterdata)"""

    record_a = record_a.copy()
    record_b = record_b.copy()

    _ensure_mandatory_fields(record_a, record_b)

    _abbreviate_container_title(record_a)
    _abbreviate_container_title(record_b)
    _format_authors_string_for_comparison(record_a)
    _format_authors_string_for_comparison(record_b)

    return _get_similarity_detailed(record_a.get_data(), record_b.get_data())


def matches(
    record_a: colrev.record.record.Record, record_b: colrev.record.record.Record
) -> bool:
    """Determine whether two records match (correspond to the same entity)."""
    record_a_dict = record_a.copy().get_data()
    record_b_dict = record_b.copy().get_data()
    record_a_dict[Fields.ID] = "a"
    record_b_dict[Fields.ID] = "b"

    records_df = pd.DataFrame([record_a_dict, record_b_dict])
    records_df = prep(records_df, verbosity_level=0, cpu=1)
    blocked_df = block(records_df, verbosity_level=0, cpu=1)
    matched_df = match(blocked_df, verbosity_level=0, cpu=1)
    duplicate_label = matched_df["duplicate_label"]
    if len(duplicate_label) == 0:  # pragma: no cover
        return False

    return duplicate_label.iloc[0] == "duplicate"



#! /usr/bin/env python
"""Completion of metadata based on year-volume-issue dependency as a prep operation"""
from __future__ import annotations

import requests
import zope.interface
from pydantic import Field

import colrev.env.local_index
import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import RecordState
from colrev.packages.crossref.src import crossref_api

# pylint: disable=duplicate-code
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-instance-attributes


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class YearVolIssPrep:
    """Prepares records based on year-volume-issue dependency"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=True)

    source_correction_hint = (
        "ask the publisher to correct the metadata"
        + " (see https://www.crossref.org/blog/"
        + "metadata-corrections-updates-and-additions-in-metadata-manager/"
    )
    always_apply_changes = True

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.prep_operation = prep_operation
        self.review_manager = prep_operation.review_manager
        self.local_index = colrev.env.local_index.LocalIndex(
            verbose_mode=self.review_manager.verbose_mode
        )
        self.vol_nr_dict = self._get_vol_nr_dict()
        self.quality_model = self.review_manager.get_qm()
        self.api = crossref_api.CrossrefAPI(params={})

    def _get_vol_nr_dict(self) -> dict:
        vol_nr_dict: dict = {}
        if not hasattr(self.review_manager, "dataset"):
            return vol_nr_dict
        records = self.review_manager.dataset.load_records_dict()
        for record in records.values():
            # pylint: disable=duplicate-code
            if record[Fields.STATUS] not in RecordState.get_post_x_states(
                state=RecordState.md_processed
            ):
                continue
            if not record.get(Fields.YEAR, "NA").isdigit():
                continue

            if Fields.JOURNAL not in record or Fields.VOLUME not in record:
                continue

            if record[Fields.JOURNAL] not in vol_nr_dict:
                vol_nr_dict[record[Fields.JOURNAL]] = {}

            if record[Fields.VOLUME] not in vol_nr_dict[record[Fields.JOURNAL]]:
                vol_nr_dict[record[Fields.JOURNAL]][record[Fields.VOLUME]] = {}

            if Fields.NUMBER not in record:
                vol_nr_dict[record[Fields.JOURNAL]][record[Fields.VOLUME]] = record[
                    Fields.YEAR
                ]
            else:
                if isinstance(
                    vol_nr_dict[record[Fields.JOURNAL]][record[Fields.VOLUME]], dict
                ):
                    vol_nr_dict[record[Fields.JOURNAL]][record[Fields.VOLUME]][
                        record[Fields.NUMBER]
                    ] = record[Fields.YEAR]
                else:
                    # do not use inconsistent data (has/has no number)
                    del vol_nr_dict[record[Fields.JOURNAL]][record[Fields.VOLUME]]

        return vol_nr_dict

    def _get_year_from_toc(self, *, record: colrev.record.record.Record) -> None:
        # TBD: maybe extract the following three lines as a separate script...
        try:
            year = self.local_index.get_year_from_toc(record.get_data())
            record.update_field(
                key=Fields.YEAR,
                value=year,
                source="LocalIndexPrep",
                note="",
                keep_source_if_equal=True,
            )
        except colrev_exceptions.TOCNotAvailableException:
            pass

    def _get_year_from_vol_nr_dict(
        self, *, record: colrev.record.record.Record
    ) -> None:
        if Fields.JOURNAL not in record.data or Fields.VOLUME not in record.data:
            return

        if record.data[Fields.JOURNAL] not in self.vol_nr_dict:
            return

        if (
            record.data[Fields.VOLUME]
            not in self.vol_nr_dict[record.data[Fields.JOURNAL]]
        ):
            return

        if Fields.NUMBER in record.data:
            if (
                record.data[Fields.NUMBER]
                in self.vol_nr_dict[record.data[Fields.JOURNAL]][
                    record.data[Fields.VOLUME]
                ]
            ):
                record.update_field(
                    key=Fields.YEAR,
                    value=self.vol_nr_dict[record.data[Fields.JOURNAL]][
                        record.data[Fields.VOLUME]
                    ][record.data[Fields.NUMBER]],
                    source="year_vol_iss_prep",
                    note="",
                )
                record.run_quality_model(self.quality_model)
        else:
            if isinstance(
                self.vol_nr_dict[record.data[Fields.JOURNAL]][
                    record.data[Fields.VOLUME]
                ],
                (str, int),
            ):
                record.update_field(
                    key=Fields.YEAR,
                    value=self.vol_nr_dict[record.data[Fields.JOURNAL]][
                        record.data[Fields.VOLUME]
                    ],
                    source="year_vol_iss_prep",
                    note="",
                )
                record.run_quality_model(self.quality_model)

    def _get_year_from_crossref(self, *, record: colrev.record.record.Record) -> None:
        try:

            retrieved_records = self.api.crossref_query(
                record_input=record,
                jour_vol_iss_list=True,
            )
            retries = 0
            while (
                not retrieved_records
                and retries < self.prep_operation.max_retries_on_error
            ):
                retries += 1
                retrieved_records = self.api.crossref_query(
                    record_input=record,
                    jour_vol_iss_list=True,
                )
            if 0 == len(retrieved_records):
                return

            retrieved_records = [
                retrieved_record
                for retrieved_record in retrieved_records
                if retrieved_record.data.get(Fields.VOLUME, "NA")
                == record.data.get(Fields.VOLUME, "NA")
                and retrieved_record.data.get(Fields.JOURNAL, "NA")
                == record.data.get(Fields.JOURNAL, "NA")
                and retrieved_record.data.get(Fields.NUMBER, "NA")
                == record.data.get(Fields.NUMBER, "NA")
            ]

            years = [r.data[Fields.YEAR] for r in retrieved_records]
            if len(years) == 0:
                return
            most_common = max(years, key=years.count)
            if years.count(most_common) > 3:
                record.update_field(
                    key=Fields.YEAR,
                    value=most_common,
                    source="CROSSREF(average)",
                    note="",
                )
        except requests.exceptions.RequestException:
            pass

    def prepare(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> colrev.record.record.Record:
        """Prepare a record based on year-volume-issue dependency"""

        if (
            record.data.get(Fields.YEAR, "NA").isdigit()
            or record.masterdata_is_curated()
        ):
            return record

        self._get_year_from_toc(record=record)

        if Fields.YEAR in record.data:
            return record

        self._get_year_from_vol_nr_dict(record=record)

        if Fields.YEAR in record.data:
            return record

        self._get_year_from_crossref(record=record)

        return record



#! /usr/bin/env python
"""Connector to OpenLibrary (API)"""
from __future__ import annotations

import json
import typing
from multiprocessing import Lock
from pathlib import Path

import requests
import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_settings
import colrev.record.record
import colrev.record.record_prep
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# Note: not (yet) implemented as a full search_source
# (including SearchSourceInterface, packages_endpoints.json)


# pylint: disable=too-few-public-methods
# pylint: disable=unused-argument


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class OpenLibrarySearchSource:
    """OpenLibrary API"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    endpoint = "colrev.open_library"
    # pylint: disable=colrev-missed-constant-usage
    source_identifier = "isbn"
    search_types = [SearchType.MD]

    ci_supported: bool = Field(default=True)
    heuristic_status = SearchSourceHeuristicStatus.na

    _open_library_md_filename = Path("data/search/md_open_library.bib")

    requests_headers = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) "
        "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36"
    }

    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager
        if settings:
            # OpenLibrary as a search_source
            self.search_source = self.settings_class(**settings)

        else:
            # OpenLibrary as an md-prep source
            open_library_md_source_l = [
                s
                for s in self.review_manager.settings.sources
                if s.filename == self._open_library_md_filename
            ]
            if open_library_md_source_l:
                self.search_source = open_library_md_source_l[0]
            else:
                self.search_source = colrev.settings.SearchSource(
                    endpoint="colrev.open_library",
                    filename=self._open_library_md_filename,
                    search_type=SearchType.MD,
                    search_parameters={},
                    comment="",
                )

            self.open_library_lock = Lock()

        self.origin_prefix = self.search_source.get_origin_prefix()

    def check_availability(
        self, *, source_operation: colrev.process.operation.Operation
    ) -> None:
        """Check the status (availability) of the OpenLibrary API"""

        test_rec = {
            Fields.ENTRYTYPE: "book",
            Fields.ISBN: "9781446201435",
            # 'author': 'Ridley, Diana',
            Fields.TITLE: "The Literature Review A Stepbystep Guide For Students",
            Fields.ID: "Ridley2012",
            Fields.YEAR: "2012",
        }
        try:
            url = f"https://openlibrary.org/isbn/{test_rec['isbn']}.json"
            ret = requests.get(
                url,
                headers=self.requests_headers,
                timeout=30,
            )
            if ret.status_code != 200:
                if not self.review_manager.force_mode:
                    raise colrev_exceptions.ServiceNotAvailableException("OPENLIBRARY")
        except requests.exceptions.RequestException as exc:
            if not self.review_manager.force_mode:
                raise colrev_exceptions.ServiceNotAvailableException(
                    "OPENLIBRARY"
                ) from exc

    # pylint: disable=colrev-missed-constant-usage
    @classmethod
    def _open_library_json_to_record(
        cls, *, item: dict, url: str
    ) -> colrev.record.record_prep.PrepRecord:
        retrieved_record: dict = {}

        if "author_name" in item:
            authors_string = " and ".join(
                [
                    colrev.record.record_prep.PrepRecord.format_author_field(author)
                    for author in item["author_name"]
                ]
            )
            retrieved_record[Fields.AUTHOR] = authors_string
        if "publisher" in item:
            retrieved_record[Fields.PUBLISHER] = str(item["publisher"][0])
        if "title" in item:
            retrieved_record[Fields.TITLE] = str(item["title"])
        if "publish_year" in item:
            retrieved_record[Fields.YEAR] = str(item["publish_year"][0])
        if "edition_count" in item:
            retrieved_record[Fields.EDITOR] = str(item["edition_count"])
        if "seed" in item:
            if "/books/" in item["seed"][0]:
                retrieved_record[Fields.ENTRYTYPE] = "book"
        if "publish_place" in item:
            retrieved_record[Fields.ADDRESS] = str(item["publish_place"][0])
        if Fields.ISBN in item:
            retrieved_record[Fields.ISBN] = str(item["isbn"][0])

        record = colrev.record.record_prep.PrepRecord(retrieved_record)
        record.add_provenance_all(source=url)
        return record

    def _get_record_from_open_library(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
    ) -> colrev.record.record.Record:
        session = prep_operation.review_manager.get_cached_session()

        url = "NA"
        if Fields.ISBN in record.data:
            isbn = record.data[Fields.ISBN].replace("-", "").replace(" ", "")
            url = f"https://openlibrary.org/isbn/{isbn}.json"
            ret = session.request(
                "GET",
                url,
                headers=self.requests_headers,
                timeout=prep_operation.timeout,
            )
            ret.raise_for_status()
            # prep_operation.review_manager.logger.debug(url)
            if '"error": "notfound"' in ret.text:
                record.remove_field(key=Fields.ISBN)

            item = json.loads(ret.text)

        else:
            base_url = "https://openlibrary.org/search.json?"
            url = ""
            if record.data.get(Fields.AUTHOR, "NA").split(",")[0]:
                url = (
                    base_url
                    + "&author="
                    + record.data.get(Fields.AUTHOR, "NA").split(",")[0]
                )
            if (
                record.data[Fields.ENTRYTYPE] == "inbook"
                and Fields.EDITOR in record.data
            ):
                if record.data.get(Fields.EDITOR, "NA").split(",")[0]:
                    url = (
                        base_url
                        + "&author="
                        + record.data.get(Fields.EDITOR, "NA").split(",")[0]
                    )
            if base_url not in url:
                raise colrev_exceptions.RecordNotFoundInPrepSourceException(
                    msg="OpenLibrary: base_url not in url"
                )

            title = record.data.get(
                Fields.TITLE, record.data.get(Fields.BOOKTITLE, "NA")
            )
            if len(title) < 10:
                raise colrev_exceptions.RecordNotFoundInPrepSourceException(
                    msg="OpenLibrary: len(title) < 10"
                )
            if ":" in title:
                title = title[: title.find(":")]  # To catch sub-titles
            url = url + "&title=" + title.replace(" ", "+")
            ret = session.request(
                "GET",
                url,
                headers=self.requests_headers,
                timeout=prep_operation.timeout,
            )
            ret.raise_for_status()
            # prep_operation.review_manager.logger.debug(url)

            # if we have an exact match, we don't need to check the similarity
            if '"numFoundExact": true,' not in ret.text:
                raise colrev_exceptions.RecordNotFoundInPrepSourceException(
                    msg="OpenLibrary: numFoundExact true missing"
                )

            data = json.loads(ret.text)
            items = data["docs"]
            if not items:
                raise colrev_exceptions.RecordNotFoundInPrepSourceException(
                    msg="OpenLibrary: no items"
                )
            item = items[0]

        retrieved_record = self._open_library_json_to_record(item=item, url=url)

        return retrieved_record

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for OpenLibrary"""

        result = {"confidence": 0.0}

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""
        raise NotImplementedError

    def search(self, rerun: bool) -> None:
        """Run a search of OpenLibrary"""

        # if self.search_source.search_type == SearchType.DB:
        #     if self.review_manager.in_ci_environment():
        #         raise colrev_exceptions.SearchNotAutomated(
        #             "DB search for OpenLibrary not automated."
        #         )

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Retrieve masterdata from OpenLibrary based on similarity with the record provided"""

        if any(self.origin_prefix in o for o in record.data[Fields.ORIGIN]):
            # Already linked to an open-library record
            return record

        try:
            retrieved_record = self._get_record_from_open_library(
                prep_operation=prep_operation, record=record
            )

            self.open_library_lock.acquire(timeout=60)
            open_library_feed = self.search_source.get_api_feed(
                review_manager=prep_operation.review_manager,
                source_identifier=self.source_identifier,
                update_only=False,
                prep_mode=True,
            )

            open_library_feed.add_update_record(retrieved_record)

            record.merge(
                retrieved_record,
                default_source=retrieved_record.data[Fields.ORIGIN][0],
            )
            open_library_feed.save()
            self.open_library_lock.release()

        except (
            colrev_exceptions.RecordNotFoundInPrepSourceException,
            requests.exceptions.RequestException,
        ):
            pass
        except (colrev_exceptions.NotFeedIdentifiableException,):
            self.open_library_lock.release()

        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for OpenLibrary"""

        return record



#!/usr/bin/env python3
"""Types and model of CoLRev operations."""
from __future__ import annotations

import typing

import docker
import git
from docker.errors import DockerException

import colrev.exceptions as colrev_exceptions
from colrev.constants import OperationsType
from colrev.process.model import ProcessModel

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.review_manager

F = typing.TypeVar("F", bound=typing.Callable[..., typing.Any])


class Operation:
    """Operations correspond to the work steps in a CoLRev project"""

    type: OperationsType

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        operations_type: OperationsType,
        notify_state_transition_operation: bool = True,
    ) -> None:
        self.review_manager = review_manager
        self.type = operations_type
        self.notify_state_transition_operation = notify_state_transition_operation
        self.notify(state_transition=notify_state_transition_operation)
        self.cpus = 4
        self.docker_images_to_stop: typing.List[str] = []

        # Note: the following call seems to block the flow (if debug is enabled)
        # self.review_manager.logger.debug(f"Created {self.type} operation")

    # pylint: disable=too-many-nested-blocks
    @classmethod
    def decorate(cls) -> typing.Callable:
        """Decorator for operations"""

        def decorator_func(func: F) -> typing.Callable:
            def wrapper_func(self, *args, **kwargs) -> typing.Any:  # type: ignore
                # Invoke the wrapped function
                retval = func(self, *args, **kwargs)
                # Conclude the operation
                self.conclude()
                if self.review_manager.in_ci_environment():
                    print("\n\n")
                return retval

            return wrapper_func

        return decorator_func

    def conclude(self) -> None:  # pragma: no cover
        """Conclude the operation (stop Docker containers)"""
        try:
            client = docker.from_env()
            for container in client.containers.list():
                if any(x in container.image.tags for x in self.docker_images_to_stop):
                    container.stop()
        except DockerException:
            pass

    def _check_model_precondition(self) -> None:
        ProcessModel.check_operation_precondition(self)

    def _require_clean_repo_general(
        self,
        *,
        ignored_files: typing.Optional[list] = None,
    ) -> bool:

        # Note : not considering untracked files.
        git_repo = git.Repo(self.review_manager.path)

        # Principle: working tree always has to be clean
        # because processing functions may change content
        if git_repo.is_dirty(index=False):
            changed_files = [item.a_path for item in git_repo.index.diff(None)]
            raise colrev_exceptions.UnstagedGitChangesError(changed_files)

        if not git_repo.is_dirty():
            return True

        ignored_file_list = [str(self.review_manager.paths.STATUS_FILE)]
        if ignored_files:
            ignored_file_list += ignored_files
        ignored_file_list = [str(x).replace("\\", "/") for x in ignored_file_list]

        changed_files = [
            item.a_path
            for item in git_repo.index.diff(None)
            if not any(ip in item.a_path.replace("\\", "/") for ip in ignored_file_list)
        ] + [
            item.a_path
            for item in git_repo.head.commit.diff()
            if not any(ip in item.a_path.replace("\\", "/") for ip in ignored_file_list)
        ]

        if changed_files:
            raise colrev_exceptions.CleanRepoRequiredError(
                changed_files, ",".join(ignored_file_list)
            )

        return True  # pragma: no cover

    def check_precondition(self) -> None:
        """Check the operation precondition"""

        if self.review_manager.force_mode:
            return

        if self.type == OperationsType.load:
            self._require_clean_repo_general(
                ignored_files=[
                    self.review_manager.paths.SEARCH_DIR,
                    self.review_manager.paths.SETTINGS_FILE,
                ]
            )
            self._check_model_precondition()

        elif self.type == OperationsType.prep_man:  # pragma: no cover
            self._require_clean_repo_general(
                ignored_files=[self.review_manager.paths.RECORDS_FILE_GIT]
            )
            self._check_model_precondition()

        elif self.type in [
            OperationsType.prep,
            OperationsType.dedupe,
            OperationsType.prescreen,
            OperationsType.pdf_prep,
            OperationsType.screen,
        ]:
            self._require_clean_repo_general()
            self._check_model_precondition()

        elif self.type in [
            OperationsType.pdf_get,
            OperationsType.pdf_get_man,
        ]:
            self._require_clean_repo_general(
                ignored_files=[self.review_manager.paths.PDF_DIR]
            )
            self._check_model_precondition()

        elif self.type == OperationsType.data:
            self._check_model_precondition()

        # ie., implicit pass for check, pdf_prep_man

    def notify(self, *, state_transition: bool = True) -> None:
        """Notify the review_manager about the next operation"""

        self.review_manager.notified_next_operation = self.type
        if state_transition:
            self.check_precondition()
        if hasattr(self.review_manager, "dataset"):
            self.review_manager.dataset.reset_log_if_no_changes()



#! /usr/bin/env python
"""Convenience functions to write files (BiBTeX, RIS, CSV, etc.)

Usage::

    TODO

"""
from __future__ import annotations

from pathlib import Path

import colrev.writer.bib
import colrev.writer.csv
import colrev.writer.excel
import colrev.writer.ris


def write_file(records_dict: dict, filename: Path, **kw) -> dict:  # type: ignore
    """Write a file (BiBTex, RIS, or other) from a dictionary of records."""
    if filename.suffix == ".bib":
        writer = colrev.writer.bib.write_file  # type: ignore
    elif filename.suffix == ".ris":
        writer = colrev.writer.ris.write_file  # type: ignore
    elif filename.suffix == ".csv":
        writer = colrev.writer.csv.write_file  # type: ignore
    elif filename.suffix == ".xlsx":
        writer = colrev.writer.excel.write_file  # type: ignore
    else:
        raise NotImplementedError

    kw["filename"] = filename
    kw["records_dict"] = records_dict

    writer(**kw)


def to_string(*, records_dict: dict, implementation: str, **kw) -> str:  # type: ignore
    """Write a string (BiBTex, RIS, or other) from a dictionary of records."""
    if implementation == "bib":
        writer = colrev.writer.bib.to_string  # type: ignore
    elif implementation == "ris":
        writer = colrev.writer.ris.to_string  # type: ignore
    else:
        raise NotImplementedError

    kw["records_dict"] = records_dict

    return writer(**kw)


# see https://github.com/MrTango/rispy/blob/main/rispy/parser.py



#! /usr/bin/env python
"""DataInterface: RefCheck"""
from zope.interface import implementer

import colrev.ops.data
import colrev.package_manager.package_settings
from colrev.package_manager.interfaces import DataInterface

# pylint: disable=unused-argument


@implementer(DataInterface)
class RefCheck:
    """RefCheck Class"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    def __init__(
        self,
        *,
        data_operation: colrev.ops.data.Data,
        settings: dict,
    ) -> None:
        self.data_operation = data_operation
        self.review_manager = data_operation.review_manager

    def update_data(
        self,
        records: dict,
        synthesized_record_status_matrix: dict,
        silent_mode: bool,
    ) -> None:
        """
        Update the data by running the data operation. This includes data extraction,
        analysis, and synthesis.

        Parameters:
        records (dict): The records to be updated.
        synthesized_record_status_matrix (dict): The status matrix for the synthesized records.
        silent_mode (bool): Whether the operation is run in silent mode
        (for checks of review_manager/status).
        """

    def update_record_status_matrix(
        self,
        synthesized_record_status_matrix: dict,
        endpoint_identifier: str,
    ) -> None:
        """Update the record status matrix,
        i.e., indicate whether the record is rev_synthesized for the given endpoint_identifier
        """

        records = self.review_manager.dataset.load_records_dict()

        for syn_id in synthesized_record_status_matrix:
            record_dict = records[syn_id]
            record = colrev.record.record.Record(record_dict)
            if record.has_quality_defects():
                synthesized_record_status_matrix[syn_id][endpoint_identifier] = False
            else:
                synthesized_record_status_matrix[syn_id][endpoint_identifier] = True

    def get_advice(
        self,
    ) -> dict:
        """Get advice on how to operate the data package endpoint"""

        advice = {
            "msg": "Data operation [ref_check endpoint]: "
            + "\n    Prepare record metadata (or mark as IGNORE:...)",
            "detailed_msg": "... with a link to the docs etc.",
        }
        return advice



#! /usr/bin/env python
"""SearchSource: Europe PMC"""
from __future__ import annotations

import json
import typing
from multiprocessing import Lock
from pathlib import Path
from sqlite3 import OperationalError
from urllib.parse import quote
from urllib.parse import urlparse

import requests
import zope.interface
from pydantic import BaseModel
from pydantic import Field
from rapidfuzz import fuzz

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
import colrev.record.record_prep
import colrev.record.record_similarity
import colrev.settings
from colrev.constants import Fields
from colrev.constants import RecordState
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.europe_pmc.src import europe_pmc_api

# pylint: disable=duplicate-code
# pylint: disable=unused-argument


class EuropePMCSearchSourceSettings(colrev.settings.SearchSource, BaseModel):
    """Settings for EuropePMCSearchSource"""

    # pylint: disable=too-many-instance-attributes
    endpoint: str
    filename: Path
    search_type: SearchType
    search_parameters: dict
    comment: typing.Optional[str]

    _details = {
        "search_parameters": {
            "tooltip": "Currently supports a scope item "
            "with venue_key and journal_abbreviated fields."
        },
    }


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class EuropePMCSearchSource:
    """Europe PMC"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    #
    source_identifier = Fields.EUROPE_PMC_ID
    search_types = [
        SearchType.API,
        SearchType.DB,
        SearchType.MD,
    ]
    endpoint = "colrev.europe_pmc"

    ci_supported: bool = Field(default=True)
    heuristic_status = SearchSourceHeuristicStatus.supported

    _europe_pmc_md_filename = Path("data/search/md_europe_pmc.bib")
    _SOURCE_URL = "https://www.ebi.ac.uk/europepmc/webservices/rest/article/"

    settings_class = EuropePMCSearchSourceSettings

    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager
        if settings:
            # EuropePMC as a search_source
            self.search_source = self.settings_class(**settings)
        else:
            # EuropePMC as an md-prep source
            europe_pmc_md_source_l = [
                s
                for s in source_operation.review_manager.settings.sources
                if s.filename == self._europe_pmc_md_filename
            ]
            if europe_pmc_md_source_l:
                self.search_source = europe_pmc_md_source_l[0]
            else:
                self.search_source = colrev.settings.SearchSource(
                    endpoint=self.endpoint,
                    filename=self._europe_pmc_md_filename,
                    search_type=SearchType.MD,
                    search_parameters={},
                    comment="",
                )

        self.europe_pmc_lock = Lock()
        self.source_operation = source_operation

    # @classmethod
    # def check_status(cls, *, prep_operation: colrev.ops.prep.Prep) -> None:
    # ...

    @classmethod
    def _get_similarity(
        cls,
        *,
        record: colrev.record.record.Record,
        retrieved_record: colrev.record.record.Record,
    ) -> float:
        title_similarity = fuzz.partial_ratio(
            retrieved_record.data[Fields.TITLE].lower(),
            record.data.get(Fields.TITLE, "").lower(),
        )
        container_similarity = fuzz.partial_ratio(
            retrieved_record.get_container_title().lower(),
            record.get_container_title().lower(),
        )

        weights = [0.6, 0.4]
        similarities = [title_similarity, container_similarity]
        similarity = sum(similarities[g] * weights[g] for g in range(len(similarities)))
        return similarity

    def _europe_pmc_query(
        self,
        *,
        record_input: colrev.record.record.Record,
        most_similar_only: bool = True,
        timeout: int = 60,
    ) -> list:
        """Retrieve records from Europe PMC based on a query"""

        try:
            api = europe_pmc_api.EPMCAPI(
                params={"query": quote(record_input.data[Fields.TITLE])},
                email=self.review_manager.get_committer()[1],
                session=self.review_manager.get_cached_session(),
            )

            record = record_input.copy_prep_rec()

            record_list = []
            counter = 0
            most_similar, most_similar_record = 0.0, {}
            for retrieved_record in api.get_records():
                counter += 1

                if Fields.TITLE not in retrieved_record.data:
                    continue

                similarity = self._get_similarity(
                    record=record, retrieved_record=retrieved_record
                )

                source = (
                    f"{self._SOURCE_URL}{retrieved_record.data[Fields.EUROPE_PMC_ID]}"
                )
                retrieved_record.set_masterdata_complete(
                    source=source,
                    masterdata_repository=self.review_manager.settings.is_curated_repo(),
                )

                if not most_similar_only:
                    record_list.append(retrieved_record)

                elif most_similar < similarity:
                    most_similar = similarity
                    most_similar_record = retrieved_record.get_data()
                if most_similar_only and counter > 5:
                    break

        except (requests.exceptions.RequestException, json.decoder.JSONDecodeError):
            return []
        except OperationalError as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                "sqlite, required for requests CachedSession "
                "(possibly caused by concurrent operations)"
            ) from exc

        if most_similar_only:
            record_list = [colrev.record.record_prep.PrepRecord(most_similar_record)]

        return record_list

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,  # pylint: disable=unused-argument
    ) -> colrev.record.record.Record:
        """Retrieve masterdata from Europe PMC based on similarity with the record provided"""

        # pylint: disable=too-many-branches
        # pylint: disable=too-many-return-statements

        # https://www.ebi.ac.uk/europepmc/webservices/rest/article/MED/23245604
        if len(record.data.get(Fields.TITLE, "")) < 35:
            return record

        try:

            retrieved_records = self._europe_pmc_query(
                record_input=record,
                timeout=timeout,
            )
            if not retrieved_records:
                return record

            retrieved_record = retrieved_records.pop()
            if 0 == len(retrieved_record.data):
                return record

            if not colrev.record.record_similarity.matches(record, retrieved_record):
                return record

            self.europe_pmc_lock.acquire(timeout=60)

            # Note : need to reload file because the object is not shared between processes
            europe_pmc_feed = self.search_source.get_api_feed(
                review_manager=prep_operation.review_manager,
                source_identifier=self.source_identifier,
                update_only=False,
                prep_mode=True,
            )
            europe_pmc_feed.add_update_record(retrieved_record=retrieved_record)

            record.merge(
                retrieved_record,
                default_source=retrieved_record.data[Fields.ORIGIN][0],
            )

            record.set_masterdata_complete(
                source=retrieved_record.data[Fields.ORIGIN][0],
                masterdata_repository=self.review_manager.settings.is_curated_repo(),
            )
            record.set_status(RecordState.md_prepared)
            europe_pmc_feed.save()

        except requests.exceptions.RequestException:
            pass
        except colrev_exceptions.NotFeedIdentifiableException:
            pass
        finally:
            try:
                self.europe_pmc_lock.release()
            except ValueError:
                pass

        return record

    def _validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""

        source = self.search_source

        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        assert source.search_type in self.search_types

        if "query" not in source.search_parameters:
            raise colrev_exceptions.InvalidQueryException(
                "Query required in search_parameters"
            )

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def search(self, rerun: bool) -> None:
        """Run a search of Europe PMC"""

        self._validate_source()
        # https://europepmc.org/RestfulWebService

        europe_pmc_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        if self.search_source.search_type == SearchType.API:
            self._run_api_search(
                europe_pmc_feed=europe_pmc_feed,
                rerun=rerun,
            )

        elif self.search_source.search_type == SearchType.DB:
            self.source_operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )

        # if self.search_source.search_type == colrev.settings.SearchSource.MD:
        # self._run_md_search_update(
        #     search_operation=search_operation,
        #     europe_pmc_feed=europe_pmc_feed,
        # )

        else:
            raise NotImplementedError

    def _run_api_search(
        self,
        *,
        europe_pmc_feed: colrev.ops.search_api_feed.SearchAPIFeed,
        rerun: bool,
    ) -> None:

        try:
            _, email = self.review_manager.get_committer()
            api = europe_pmc_api.EPMCAPI(
                params=self.search_source.search_parameters,
                email=email,
                session=self.review_manager.get_cached_session(),
            )

            while api.url:

                for retrieved_record in api.get_records():
                    if Fields.TITLE not in retrieved_record.data:
                        self.review_manager.logger.warning(
                            f"Skipped record: {retrieved_record.data}"
                        )
                        continue

                    source = f"{self._SOURCE_URL}{retrieved_record.data[Fields.EUROPE_PMC_ID]}"
                    retrieved_record.set_masterdata_complete(
                        source=source,
                        masterdata_repository=self.review_manager.settings.is_curated_repo(),
                    )

                    europe_pmc_feed.add_update_record(retrieved_record)

        except (requests.exceptions.RequestException, json.decoder.JSONDecodeError):
            pass
        except OperationalError as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                "sqlite, required for requests CachedSession "
                "(possibly caused by concurrent operations)"
            ) from exc
        finally:
            europe_pmc_feed.save()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Europe PMC"""

        result = {"confidence": 0.0}
        # pylint: disable=colrev-missed-constant-usage
        if "europe_pmc_id" in data:
            result["confidence"] = 1.0

        if "https://europepmc.org" in data:  # nosec
            if data.count("https://europepmc.org") >= data.count("\n@"):
                result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        params_dict = {}
        if params:
            if params.startswith("http"):
                params_dict = {Fields.URL: params}
            else:
                for item in params.split(";"):
                    key, value = item.split("=")
                    params_dict[key] = value

        if len(params_dict) == 0:
            search_source = operation.create_api_source(endpoint=cls.endpoint)

        # pylint: disable=colrev-missed-constant-usage
        elif "url" in params_dict:
            host = urlparse(params_dict["url"]).hostname

            if host and host.endswith("europepmc.org"):
                query = params_dict["url"].replace(
                    "https://europepmc.org/search?query=", ""
                )
                filename = operation.get_unique_filename(file_path_string="europepmc")
                search_source = colrev.settings.SearchSource(
                    endpoint=cls.endpoint,
                    filename=filename,
                    search_type=SearchType.API,
                    search_parameters={"query": query},
                    comment="",
                )
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError

        operation.add_source_and_search(search_source)
        return search_source

    def _load_bib(self) -> dict:
        def field_mapper(record_dict: dict) -> None:
            for key in list(record_dict.keys()):
                if key not in ["ID", "ENTRYTYPE"]:
                    record_dict[key.lower()] = record_dict.pop(key)

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            logger=self.review_manager.logger,
            unique_id_field="ID",
            field_mapper=field_mapper,
        )
        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            return self._load_bib()

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for Europe PMC"""
        record.data[Fields.AUTHOR].rstrip(".")
        record.data[Fields.TITLE].rstrip(".")
        return record



#! /usr/bin/env python
"""Deduplication of remaining records in curated metadata repositories"""
from __future__ import annotations

import typing
from pathlib import Path

import pandas as pd
import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
import colrev.record.record_prep
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import RecordState


# pylint: disable=too-many-arguments
# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.DedupeInterface)
class CurationMissingDedupe:
    """Deduplication of remaining records in a curated metadata repository"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=False)

    def __init__(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,
        settings: dict,
    ):
        self.settings = self.settings_class(**settings)
        self.review_manager = dedupe_operation.review_manager
        self.dedupe_operation = dedupe_operation

        self._post_md_prepared_states = RecordState.get_post_x_states(
            state=RecordState.md_processed
        )

    def _create_dedupe_source_stats(self) -> None:
        # Note : reload to generate correct statistics

        Path("dedupe").mkdir(exist_ok=True)

        source_origins = [
            str(source.filename).replace("data/search/", "")
            for source in self.review_manager.settings.sources
        ]

        records = self.review_manager.dataset.load_records_dict()
        for source_origin in source_origins:
            selected_records = [
                r
                for r in records.values()
                if any(source_origin in co for co in r[Fields.ORIGIN])
                and r[Fields.STATUS]
                in [
                    RecordState.md_prepared,
                    RecordState.md_needs_manual_preparation,
                    RecordState.md_imported,
                ]
            ]
            records_df = pd.DataFrame.from_records(list(selected_records))
            if records_df.shape[0] == 0:
                self.review_manager.logger.info(
                    f"{Colors.GREEN}Source {source_origin} fully merged{Colors.END}"
                )
            else:
                self.review_manager.logger.info(
                    f"{Colors.ORANGE}Source {source_origin} not fully merged{Colors.END}"
                )
                self.review_manager.logger.info(
                    f"Exporting details to dedupe/{source_origin}.xlsx"
                )

                records_df = records_df[
                    records_df.columns.intersection(
                        [
                            Fields.ID,
                            Fields.STATUS,
                            Fields.JOURNAL,
                            Fields.BOOKTITLE,
                            Fields.YEAR,
                            Fields.VOLUME,
                            Fields.NUMBER,
                            Fields.TITLE,
                            Fields.AUTHOR,
                        ]
                    )
                ]
                keys = list(
                    records_df.columns.intersection(
                        [Fields.YEAR, Fields.VOLUME, Fields.NUMBER]
                    )
                )
                if Fields.YEAR in keys:
                    records_df.year = pd.to_numeric(records_df.year, errors="coerce")
                if Fields.VOLUME in keys:
                    records_df.volume = pd.to_numeric(
                        records_df.volume, errors="coerce"
                    )
                if Fields.NUMBER in keys:
                    records_df.number = pd.to_numeric(
                        records_df.number, errors="coerce"
                    )
                records_df.sort_values(by=keys, inplace=True)
                records_df.to_excel(f"dedupe/{source_origin}.xlsx", index=False)

    def _get_same_toc_recs(
        self, *, record: colrev.record.record.Record, records: dict
    ) -> list:
        if self.review_manager.force_mode:
            if record.data[Fields.STATUS] in self._post_md_prepared_states:
                return []
        else:
            # only dedupe md_prepared records
            if record.data[Fields.STATUS] not in [RecordState.md_prepared]:
                return []
        if record.data.get(Fields.TITLE, "") == "":
            return []

        try:
            toc_key = record.get_toc_key()
        except colrev_exceptions.NotTOCIdentifiableException:
            return []

        same_toc_recs = []
        for record_candidate in records.values():
            try:
                candidate_toc_key = colrev.record.record.Record(
                    record_candidate
                ).get_toc_key()
            except colrev_exceptions.NotTOCIdentifiableException:
                continue
            if toc_key != candidate_toc_key:
                continue
            if record_candidate[Fields.ID] == record.data[Fields.ID]:
                continue
            if record_candidate[Fields.STATUS] in [
                RecordState.md_prepared,
                RecordState.md_needs_manual_preparation,
                RecordState.md_imported,
            ]:
                continue
            same_toc_recs.append(record_candidate)
        return same_toc_recs

    def _print_same_toc_recs(
        self, *, same_toc_recs: list, record: colrev.record.record.Record
    ) -> list:
        for same_toc_rec in same_toc_recs:
            same_toc_rec["similarity"] = (
                colrev.record.record_prep.PrepRecord.get_record_similarity(
                    colrev.record.record.Record(same_toc_rec), record
                )
            )

        same_toc_recs = sorted(
            same_toc_recs, key=lambda d: d["similarity"], reverse=True
        )
        if len(same_toc_recs) > 20:
            same_toc_recs = same_toc_recs[0:20]

        i = 0
        for i, same_toc_rec in enumerate(same_toc_recs):
            author_title_string = (
                f"{same_toc_rec.get('author', 'NO_AUTHOR')} : "
                + f"{same_toc_rec.get('title', 'NO_TITLE')}"
            )

            if same_toc_rec["similarity"] > 0.8:
                print(f"{i + 1} - {Colors.ORANGE}{author_title_string}{Colors.END}")

            else:
                print(f"{i + 1} - {author_title_string}")
        return same_toc_recs

    def _get_nr_recs_to_merge(self, *, records: dict) -> int:
        if self.review_manager.force_mode:
            self.review_manager.logger.info(
                "Scope: md_prepared, md_needs_manual_preparation, md_imported"
            )
            nr_recs_to_merge = len(
                [
                    x
                    for x in records.values()
                    if x[Fields.STATUS] not in self._post_md_prepared_states
                ]
            )
        else:
            self.review_manager.logger.info("Scope: md_prepared")
            nr_recs_to_merge = len(
                [
                    x
                    for x in records.values()
                    if x[Fields.STATUS] in [RecordState.md_prepared]
                ]
            )
        return nr_recs_to_merge

    def _process_missing_duplicates(self) -> dict:
        records = self.review_manager.dataset.load_records_dict()

        nr_recs_to_merge = self._get_nr_recs_to_merge(records=records)

        nr_recs_checked = 0
        results: typing.Dict[str, list] = {
            "decision_list": [],
            "add_records_to_md_processed_list": [],
            "records_to_prepare": [],
        }

        for record_dict in records.values():
            record = colrev.record.record.Record(record_dict)

            same_toc_recs = self._get_same_toc_recs(record=record, records=records)

            if len(same_toc_recs) == 0:
                print("no same toc records")
                continue

            print("\n\n\n")
            print(Colors.ORANGE)
            record.print_citation_format()
            print(Colors.END)

            same_toc_recs = self._print_same_toc_recs(
                same_toc_recs=same_toc_recs, record=record
            )
            i = len(same_toc_recs)
            valid_selection = False
            quit_pressed = False
            while not valid_selection:
                ret = input(
                    f"({nr_recs_checked}/{nr_recs_to_merge}) "
                    f"Merge with record [{1}...{i+1} / s / a / p / q]?   "
                )
                if ret == "s":
                    valid_selection = True
                elif ret == "q":
                    quit_pressed = True
                    valid_selection = True
                elif ret == "a":
                    results["add_records_to_md_processed_list"].append(
                        record.data[Fields.ID]
                    )
                    valid_selection = True
                elif ret == "p":
                    results["records_to_prepare"].append(record.data[Fields.ID])
                    valid_selection = True
                elif ret.isdigit():
                    if int(ret) - 1 <= i:
                        rec2 = same_toc_recs[int(ret) - 1]
                        if record.data[Fields.STATUS] < rec2[Fields.STATUS]:
                            results["decision_list"].append(
                                [rec2[Fields.ID], record.data[Fields.ID]]
                            )
                        else:
                            results["decision_list"].append(
                                [rec2[Fields.ID], record.data[Fields.ID]]
                            )

                        valid_selection = True
            nr_recs_checked += 1
            if quit_pressed:
                break
        return results

    def run_dedupe(self) -> None:
        """Run the dedupe procedure for remaining records in curations"""

        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements

        # export sets of non-merged records
        # (and merged records a different xlsx for easy sort/merge)

        # Note : this script is necessary because the active learning is insufficient:
        # the automated ML-deduplication still has a certain error rate
        # which makes it less useful for curations
        # the active learning labeling presents cases on both sides
        # (likely duplicates and non-duplicates to maximize training quality)
        # For the curation, we are only interested in the duplicate, not the classifier

        print("\n\n")
        print(
            "In the following, "
            "records can be added to the curated (md_processed*) records.\n"
            "Curated records are displayed for the same table-of-content item "
            "(i.e., same year/volume/number)"
        )
        print("\n\n")

        ret = self._process_missing_duplicates()

        if len(ret["decision_list"]) > 0:
            print("Duplicates identified:")
            print(ret["decision_list"])
            preferred_masterdata_sources = [
                s
                for s in self.review_manager.settings.sources
                if s.endpoint != "colrev.files_dir"
            ]

            self.dedupe_operation.apply_merges(
                id_sets=ret["decision_list"],
                preferred_masterdata_sources=preferred_masterdata_sources,
            )

        if len(ret["records_to_prepare"]) > 0:
            records = self.review_manager.dataset.load_records_dict()
            for record_id, record_dict in records.items():
                if record_id in ret["records_to_prepare"]:
                    record = colrev.record.record.Record(record_dict)
                    record.set_status(
                        target_state=RecordState.md_needs_manual_preparation
                    )

            self.review_manager.dataset.save_records_dict(records)

        if len(ret["decision_list"]) > 0 or len(ret["records_to_prepare"]) > 0:
            self.review_manager.dataset.create_commit(
                msg="Merge duplicate records",
            )

        if len(ret["add_records_to_md_processed_list"]) > 0:
            records = self.review_manager.dataset.load_records_dict()
            for record_id, record_dict in records.items():
                if record_id in ret["add_records_to_md_processed_list"]:
                    if record_dict[Fields.STATUS] in [
                        RecordState.md_prepared,
                        RecordState.md_needs_manual_preparation,
                        RecordState.md_imported,
                    ]:
                        record = colrev.record.record.Record(record_dict)
                        record.set_status(RecordState.md_processed)

            self.review_manager.dataset.save_records_dict(records)
            input("Edit records (if any), add to git, and press Enter")

            self.review_manager.dataset.create_commit(
                msg="Add non-duplicate records",
            )

        self._create_dedupe_source_stats()



#! /usr/bin/env python
"""Conditional prescreen"""
from __future__ import annotations

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import RecordState


# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.PrescreenInterface)
class ConditionalPrescreen:
    """Conditional prescreen (currently: include all)"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=True)

    def __init__(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.review_manager = prescreen_operation.review_manager

    def run_prescreen(
        self,
        records: dict,
        split: list,  # pylint: disable=unused-argument
    ) -> dict:
        """Prescreen records based on predefined conditions (rules)"""

        pad = 50
        for record in records.values():
            if record[Fields.STATUS] != RecordState.md_processed:
                continue
            self.review_manager.report_logger.info(
                f" {record[Fields.ID]}".ljust(pad, " ")
                + "Included in prescreen (automatically)"
            )
            # pylint: disable=colrev-direct-status-assign
            record.update(colrev_status=RecordState.rev_prescreen_included)

        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.dataset.create_commit(
            msg="Prescreen: include all",
            manual_author=False,
        )
        return records



#! /usr/bin/env python
"""API for Unpaywall"""
from __future__ import annotations

import re
import typing

import requests

import colrev.exceptions as colrev_exceptions
import colrev.record.record
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.env.environment_manager import EnvironmentManager
from colrev.packages.unpaywall.src import utils


ENTRYTYPE_MAPPING = {
    "journal-article": ENTRYTYPES.ARTICLE,
    "book": ENTRYTYPES.BOOK,
    "proceedings-article": ENTRYTYPES.INPROCEEDINGS,
    "book-chapter": ENTRYTYPES.INBOOK,
    "conference": ENTRYTYPES.CONFERENCE,
    "dissertation": ENTRYTYPES.PHDTHESIS,
    "report": ENTRYTYPES.TECHREPORT,
    "other": ENTRYTYPES.MISC,
    "book-section": ENTRYTYPES.INBOOK,
    "monograph": ENTRYTYPES.THESIS,
    "report-component": ENTRYTYPES.TECHREPORT,
    "peer-review": ENTRYTYPES.MISC,
    "book-track": ENTRYTYPES.INCOLLECTION,
    "book-part": ENTRYTYPES.INBOOK,
    "journal-volume": ENTRYTYPES.ARTICLE,
    "book-set": ENTRYTYPES.MISC,
    "reference-entry": ENTRYTYPES.MISC,
    "journal": ENTRYTYPES.MISC,
    "component": ENTRYTYPES.MISC,
    "proceedings-series": ENTRYTYPES.PROCEEDINGS,
    "report-series": ENTRYTYPES.TECHREPORT,
    "proceedings": ENTRYTYPES.PROCEEDINGS,
    "database": ENTRYTYPES.MISC,
    "standard": ENTRYTYPES.MISC,
    "reference-book": ENTRYTYPES.BOOK,
    "posted-content": ENTRYTYPES.MISC,
    "journal-issue": ENTRYTYPES.MISC,
    "grant": ENTRYTYPES.MISC,
    "dataset": ENTRYTYPES.MISC,
    "book-series": ENTRYTYPES.BOOK,
    "edited-book": ENTRYTYPES.BOOK,
}


class UnpaywallAPI:
    """API for Unpaywall"""

    def __init__(self, search_parameters: dict) -> None:
        self.search_parameters = search_parameters

    def _get_authors(self, article: dict) -> typing.List[str]:
        authors = []
        z_authors = article.get("z_authors", [])
        if z_authors:
            for author in z_authors:
                given_name = author.get("given", "")
                family_name = author.get("family", "")
                if given_name and family_name:
                    authors.append(f"{family_name}, {given_name}")
        return authors

    def _get_affiliation(self, article: dict) -> typing.List[str]:
        affiliations = set()
        z_authors = article.get("z_authors", "")
        if z_authors:
            for person in z_authors:
                person_affiliation = person.get("affiliation", [])
                if person_affiliation:
                    affiliations.add(person_affiliation[0]["name"])

        return list(affiliations)

    def create_record(self, article: dict) -> colrev.record.record.Record:
        """Build record"""
        doi = article.get("doi", "").upper()
        record_dict = {Fields.ID: doi}

        entrytype = ENTRYTYPE_MAPPING.get(
            article.get("genre", "other"), ENTRYTYPES.MISC
        )
        record_dict[Fields.ENTRYTYPE] = entrytype

        record_dict[Fields.AUTHOR] = " and ".join(self._get_authors(article))
        record_dict[Fields.TITLE] = article.get("title", "")
        record_dict[Fields.YEAR] = article.get("year", "")
        record_dict[Fields.DOI] = doi

        if entrytype == ENTRYTYPES.ARTICLE:
            record_dict[Fields.JOURNAL] = article.get("journal_name", "")
        elif entrytype == ENTRYTYPES.BOOK:
            record_dict[Fields.PUBLISHER] = article.get("publisher", "")
        elif entrytype == ENTRYTYPES.INPROCEEDINGS:
            record_dict[Fields.BOOKTITLE] = article.get("journal_name", "")
        elif entrytype == ENTRYTYPES.INBOOK:
            record_dict[Fields.BOOKTITLE] = article.get("journal_name", "")
            record_dict[Fields.PUBLISHER] = article.get("publisher", "")
        elif entrytype == ENTRYTYPES.CONFERENCE:
            record_dict[Fields.BOOKTITLE] = article.get("journal_name", "")
        elif entrytype == ENTRYTYPES.PHDTHESIS:
            record_dict[Fields.SCHOOL] = ",".join(self._get_affiliation(article))
        elif entrytype == ENTRYTYPES.TECHREPORT:
            record_dict[Fields.INSTITUTION] = ",".join(self._get_affiliation(article))
        elif entrytype == ENTRYTYPES.INCOLLECTION:
            record_dict[Fields.BOOKTITLE] = article.get("journal_name", "")
            record_dict[Fields.PUBLISHER] = article.get("publisher", "")

        bestoa = article.get("best_oa_location", "")
        if bestoa:
            record_dict[Fields.URL] = bestoa.get("url_for_landing_page", "")
            record_dict[Fields.FULLTEXT] = bestoa.get("url_for_pdf", "")

        final_record_dict = {key: value for key, value in record_dict.items() if value}

        record = colrev.record.record.Record(final_record_dict)

        return record

    def get_query_records(self) -> typing.Iterator[colrev.record.record.Record]:
        """Get the records from a query"""
        all_results = []
        page = 1
        results_per_page = 50

        while True:
            page_dependend_url = self._build_search_url(page)
            response = requests.get(page_dependend_url, timeout=90)
            if response.status_code != 200:
                print(f"Error fetching data: {response.status_code}")
                return

            data = response.json()
            if "results" not in data:
                raise colrev_exceptions.ServiceNotAvailableException(
                    f"Could not reach API. Status Code: {response.status_code}"
                )

            new_results = data["results"]
            for result in new_results:
                if result not in all_results:
                    all_results.append(result)

            if len(new_results) < results_per_page:
                break

            page += 1

        for result in all_results:
            article = result["response"]
            record = self.create_record(article)
            yield record

    def _build_search_url(self, page: int) -> str:
        url = "https://api.unpaywall.org/v2/search?"
        params = self.search_parameters
        query = self._encode_query_for_html_url(params["query"])
        is_oa = params.get("is_oa", "null")
        email_param = params.get("email", "")

        if email_param and page == 1:

            env_man = EnvironmentManager()
            path = utils.UNPAYWALL_EMAIL_PATH
            value_string = email_param
            print(f"Updating registry settings:\n{path} = {value_string}")
            env_man.update_registry(path, value_string)

        email = utils.get_email()

        return f"{url}query={query}&is_oa={is_oa}&page={page}&email={email}"

    @classmethod
    def decode_html_url_encoding_to_string(cls, query: str) -> str:
        """Decode URL encoding to string"""
        query = query.replace("AND", "%20")
        query = re.sub(r"(%20)+", "%20", query).strip()
        query = query.replace("%20OR%20", " OR ")
        query = query.replace("%20-", " NOT ")
        query = query.replace(" -", " NOT ")
        query = query.replace("%20", " AND ")
        query = re.sub(r"\s+", " ", query).strip()
        query = query.lstrip(" ")
        query = query.rstrip(" ")
        query = query.replace("%22", '"')
        return query

    def _encode_query_for_html_url(self, query: str) -> str:
        query = query.replace("'", '"')
        query = re.sub(r"\s+", " ", query).strip()
        splited_query = query.split(" ")
        is_in_quotes = False
        parts = []
        for query_part in splited_query:
            if not is_in_quotes and query_part.startswith('"'):
                parts.append(query_part)
                is_in_quotes = True
            elif is_in_quotes and query_part.endswith('"'):
                parts.append(query_part)
                is_in_quotes = False
            elif is_in_quotes:
                parts.append(query_part)
            else:
                query_part = query_part.replace("OR", "%20OR%20")
                query_part = query_part.replace("NOT", "%20-")
                query_part = query_part.replace("AND", "%20")
                query_part = query_part.replace(" ", "%20")
                parts.append(query_part)
        joined_query = "%20".join(parts)
        joined_query = re.sub(r"(%20)+", "%20", joined_query).strip()
        joined_query = joined_query.replace("-%20", "-")
        return joined_query



#! /usr/bin/env python
"""SearchSource: ABI/INFORM (ProQuest)"""
from __future__ import annotations

import re
from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.writer.write_utils import write_file

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class ABIInformProQuestSearchSource:
    """ABI/INFORM (ProQuest)"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    endpoint = "colrev.abi_inform_proquest"
    source_identifier = "{{ID}}"
    search_types = [SearchType.DB]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported

    db_url = "https://search.proquest.com/abicomplete/advanced"

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.review_manager = source_operation.review_manager
        self.search_source = self.settings_class(**settings)
        self.source_operation = source_operation
        self.quality_model = self.review_manager.get_qm()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for ABI/INFORM (ProQuest)"""

        result = {"confidence": 0.0}

        if "proquest.com" in data:  # nosec
            if data.count("proquest.com") >= data.count("\n@"):
                result["confidence"] = 1.0
            if data.count("proquest.com") >= data.count("TY  -"):
                result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint"""

        search_source = operation.create_db_source(
            search_source_cls=cls,
            params={},
        )
        operation.add_source_and_search(search_source)
        return search_source

    def search(self, rerun: bool) -> None:
        """Run a search of ABI/INFORM"""

        if self.search_source.search_type == SearchType.DB:
            self.source_operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )

    def _remove_duplicates(self, records: dict) -> None:
        to_delete = []
        for record in records.values():
            if re.search(r"-\d{1,2}$", record[Fields.ID]):
                original_record_id = re.sub(r"-\d{1,2}$", "", record[Fields.ID])
                if original_record_id not in records:
                    continue
                original_record = records[original_record_id]

                # Note: between duplicate records,
                # there are variations in spelling and completeness
                if (
                    colrev.record.record.Record.get_record_similarity(
                        record_a=colrev.record.record.Record(record),
                        record_b=colrev.record.record.Record(original_record),
                    )
                    < 0.9
                ):
                    continue

                if original_record_id not in records:
                    continue
                to_delete.append(record[Fields.ID])
        if to_delete:
            for rid in to_delete:
                self.review_manager.logger.info(f" remove duplicate {rid}")
                del records[rid]

            write_file(records_dict=records, filename=self.search_source.filename)

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def _load_ris(self) -> dict:

        def id_labeler(records: list) -> None:
            for record_dict in records:
                record_dict[Fields.ID] = record_dict["AN"]

        def entrytype_setter(record_dict: dict) -> None:
            if record_dict["TY"] == "JOUR":
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
            elif record_dict["TY"] == "BOOK":
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.BOOK
            elif record_dict["TY"] == "THES":
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.PHDTHESIS
            else:
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC

        def field_mapper(record_dict: dict) -> None:

            key_maps = {
                ENTRYTYPES.ARTICLE: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "TI": Fields.TITLE,
                    "JF": Fields.JOURNAL,
                    "AB": Fields.ABSTRACT,
                    "VL": Fields.VOLUME,
                    "IS": Fields.NUMBER,
                    "KW": Fields.KEYWORDS,
                    "DO": Fields.DOI,
                    "PB": Fields.PUBLISHER,
                    "SP": Fields.PAGES,
                    "PMID": Fields.PUBMED_ID,
                    "SN": Fields.ISSN,
                    "AN": f"{self.endpoint}.accession_number",
                    "LA": Fields.LANGUAGE,
                    "L2": Fields.FULLTEXT,
                    "UR": Fields.URL,
                },
                ENTRYTYPES.PHDTHESIS: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "T1": Fields.TITLE,
                    "UR": Fields.URL,
                    "PB": Fields.SCHOOL,
                    "KW": Fields.KEYWORDS,
                    "AN": f"{self.endpoint}.accession_number",
                    "AB": Fields.ABSTRACT,
                    "LA": Fields.LANGUAGE,
                    "CY": Fields.ADDRESS,
                    "L2": Fields.FULLTEXT,
                    "A3": f"{self.endpoint}.supervisor",
                },
            }

            if record_dict[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
                if "T1" in record_dict and "TI" not in record_dict:
                    record_dict["TI"] = record_dict.pop("T1")

            key_map = key_maps[record_dict[Fields.ENTRYTYPE]]
            for ris_key in list(record_dict.keys()):
                if ris_key in key_map:
                    standard_key = key_map[ris_key]
                    record_dict[standard_key] = record_dict.pop(ris_key)

            if "SP" in record_dict and "EP" in record_dict:
                record_dict[Fields.PAGES] = (
                    f"{record_dict.pop('SP')}--{record_dict.pop('EP')}"
                )

            if Fields.AUTHOR in record_dict and isinstance(
                record_dict[Fields.AUTHOR], list
            ):
                record_dict[Fields.AUTHOR] = " and ".join(record_dict[Fields.AUTHOR])
            if Fields.EDITOR in record_dict and isinstance(
                record_dict[Fields.EDITOR], list
            ):
                record_dict[Fields.EDITOR] = " and ".join(record_dict[Fields.EDITOR])
            if Fields.KEYWORDS in record_dict and isinstance(
                record_dict[Fields.KEYWORDS], list
            ):
                record_dict[Fields.KEYWORDS] = ", ".join(record_dict[Fields.KEYWORDS])

            keys_to_remove = [
                "TY",
                "Y2",
                "DB",
                "C1",
                "T3",
                "DA",
                "JF",
                "L1",
                "SP",
                "Y1",
                "M1",
                "M3",
                "N1",
                "PP",
                "CY",
                "SN",
                "ER",
                "AN",
            ]

            for key in keys_to_remove:
                record_dict.pop(key, None)

            for key, value in record_dict.items():
                record_dict[key] = str(value)

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            id_labeler=id_labeler,
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=self.review_manager.logger,
        )

        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
                unique_id_field="ID",
            )
            self._remove_duplicates(records)
            return records

        if self.search_source.filename.suffix == ".ris":
            return self._load_ris()

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for ABI/INFORM (ProQuest)"""

        if (
            record.data.get(Fields.JOURNAL, "")
            .lower()
            .endswith("conference proceedings.")
        ):
            record.change_entrytype(
                new_entrytype="inproceedings", qm=self.quality_model
            )

        if Fields.LANGUAGE in record.data:
            if record.data[Fields.LANGUAGE] in ["ENG", "English"]:
                record.update_field(
                    key=Fields.LANGUAGE,
                    value="eng",
                    source="prep_abi_inform_proquest_source",
                )

        return record



#! /usr/bin/env python
"""Source-specific preparation as a prep operation"""
from __future__ import annotations

from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import EndpointType
from colrev.constants import Fields

# pylint: disable=duplicate-code


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class SourceSpecificPrep:
    """Prepares records based on the prepare scripts specified by the SearchSource"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    source_correction_hint = "check with the developer"
    ci_supported: bool = Field(default=True)

    always_apply_changes = True

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.review_manager = prep_operation.review_manager

        self.package_manager = prep_operation.review_manager.get_package_manager()

    def prepare(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> colrev.record.record.Record:
        """Prepare the record by applying source-specific fixes"""

        # Note : we take the first origin (ie., the source-specific prep should
        # be one of the first in the prep-list)
        origin_source = record.data[Fields.ORIGIN][0].split("/")[0]

        sources = [
            s
            for s in self.review_manager.settings.sources
            if s.filename == Path("data/search") / Path(origin_source)
        ]

        for source in sources:
            try:
                # if source.endpoint not in self.search_sources.packages:
                #     continue
                # endpoint = self.search_sources.packages[source.endpoint]
                search_source_class = self.package_manager.get_package_endpoint_class(
                    package_type=EndpointType.search_source,
                    package_identifier=source.endpoint,
                )
                endpoint = search_source_class(
                    source_operation=self, settings=source.model_dump()
                )

                if callable(endpoint.prepare):
                    record = endpoint.prepare(record, source)
                else:
                    print(f"error: {source.endpoint}")
            except colrev_exceptions.MissingDependencyError as exc:
                self.review_manager.logger.warn(exc)

        if "howpublished" in record.data and Fields.URL not in record.data:
            if Fields.URL in record.data["howpublished"]:
                record.rename_field(key="howpublished", new_key=Fields.URL)
                record.update_field(
                    key=Fields.URL,
                    value=record.data[Fields.URL].replace("\\url{", "").rstrip("}"),
                    source="source_specific_prep",
                )

        if "webpage" == record.data[Fields.ENTRYTYPE].lower() or (
            "misc" == record.data[Fields.ENTRYTYPE].lower()
            and Fields.URL in record.data
        ):
            record.update_field(
                key=Fields.ENTRYTYPE, value="online", source="source_specific_prep"
            )

        return record



#! /usr/bin/env python
"""Manages Docker"""
from __future__ import annotations

import typing
from pathlib import Path

import docker
from docker.errors import DockerException

import colrev.exceptions as colrev_exceptions
from colrev.constants import Colors


class DockerManager:
    """The DockerManager manages everything related to Docker
    (e.g. building images, running containers)"""

    @classmethod
    def build_docker_image(
        cls, *, imagename: str, dockerfile: typing.Optional[Path] = None
    ) -> None:
        """Build a docker image"""

        try:
            client = docker.from_env()
            repo_tags = [t for image in client.images.list() for t in image.tags]

            if imagename not in repo_tags:
                if dockerfile:
                    if ":" not in imagename:
                        tag = f"{imagename}:latest"
                    else:
                        tag = imagename
                    dockerfile.resolve()
                    client.images.build(
                        path=str(dockerfile.parent).replace("\\", "/"),
                        tag=tag,
                        rm=True,
                    )

                else:
                    print(f"Pulling {imagename} Docker image...")
                    client.images.pull(imagename)
        except DockerException as exc:  # pragma: no cover
            raise colrev_exceptions.ServiceNotAvailableException(
                dep="docker",
                detailed_trace=f"Docker service not available ({exc}). "
                + "Please install/start Docker.",
            ) from exc

    @classmethod
    def check_docker_installed(cls) -> None:  # pragma: no cover
        """Check whether Docker is installed"""

        try:
            client = docker.from_env()
            _ = client.version()
        except docker.errors.DockerException as exc:
            if "PermissionError" in exc.args[0]:
                raise colrev_exceptions.DependencyConfigurationError(
                    "Docker: Permission error. Run "
                    f"{Colors.ORANGE}sudo gpasswd -a $USER docker && "
                    f"newgrp docker{Colors.END}"
                )
            raise colrev_exceptions.MissingDependencyError("Docker")



#! /usr/bin/env python
"""SearchSource: JSTOR"""
from __future__ import annotations

from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class JSTORSearchSource:
    """JSTOR"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    endpoint = "colrev.jstor"
    # pylint: disable=colrev-missed-constant-usage
    source_identifier = "url"
    search_types = [SearchType.DB]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported

    db_url = "http://www.jstor.org"

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.search_source = self.settings_class(**settings)
        self.operation = source_operation
        self.review_manager = source_operation.review_manager

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for JSTOR"""

        result = {"confidence": 0.1}

        if "www.jstor.org:" in data:
            if data.count("www.jstor.org") > data.count("\n@"):
                result["confidence"] = 1.0
        if data.startswith("Provider: JSTOR http://www.jstor.org"):
            result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        search_source = operation.create_db_source(
            search_source_cls=cls,
            params={},
        )
        operation.add_source_and_search(search_source)
        return search_source

    def search(self, rerun: bool) -> None:
        """Run a search of JSTOR"""

        if self.search_source.search_type == SearchType.DB:
            self.operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )
            return

        raise NotImplementedError

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def _load_ris(self) -> dict:

        def id_labeler(records: list) -> None:
            for record_dict in records:
                record_dict[Fields.ID] = record_dict["UR"].split("/")[-1]

        def entrytype_setter(record_dict: dict) -> None:
            if record_dict["TY"] == "JOUR":
                record_dict[Fields.ENTRYTYPE] = "article"
            elif record_dict["TY"] == "RPRT":
                record_dict[Fields.ENTRYTYPE] = "techreport"
            elif record_dict["TY"] == "CHAP":
                record_dict[Fields.ENTRYTYPE] = "inbook"
            else:
                record_dict[Fields.ENTRYTYPE] = "misc"

        def field_mapper(record_dict: dict) -> None:

            key_maps = {
                ENTRYTYPES.ARTICLE: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "TI": Fields.TITLE,
                    "T2": Fields.JOURNAL,
                    "AB": Fields.ABSTRACT,
                    "VL": Fields.VOLUME,
                    "IS": Fields.NUMBER,
                    "DO": Fields.DOI,
                    "PB": Fields.PUBLISHER,
                    "UR": Fields.URL,
                    "SN": Fields.ISSN,
                },
                ENTRYTYPES.INBOOK: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "TI": Fields.CHAPTER,
                    "T2": Fields.TITLE,
                    "DO": Fields.DOI,
                    "PB": Fields.PUBLISHER,
                    "UR": Fields.URL,
                    "AB": Fields.ABSTRACT,
                    "SN": Fields.ISBN,
                    "A2": Fields.EDITOR,
                },
                ENTRYTYPES.TECHREPORT: {
                    "PY": Fields.YEAR,
                    "AU": Fields.AUTHOR,
                    "TI": Fields.TITLE,
                    "UR": Fields.URL,
                    "PB": Fields.PUBLISHER,
                },
            }

            if record_dict[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
                if "T1" in record_dict and "TI" not in record_dict:
                    record_dict["TI"] = record_dict.pop("T1")

            key_map = key_maps[record_dict[Fields.ENTRYTYPE]]
            for ris_key in list(record_dict.keys()):
                if ris_key in key_map:
                    standard_key = key_map[ris_key]
                    record_dict[standard_key] = record_dict.pop(ris_key)

            if "SP" in record_dict and "EP" in record_dict:
                record_dict[Fields.PAGES] = (
                    f"{record_dict.pop('SP')}--{record_dict.pop('EP')}"
                )

            if Fields.AUTHOR in record_dict and isinstance(
                record_dict[Fields.AUTHOR], list
            ):
                record_dict[Fields.AUTHOR] = " and ".join(record_dict[Fields.AUTHOR])
            if Fields.EDITOR in record_dict and isinstance(
                record_dict[Fields.EDITOR], list
            ):
                record_dict[Fields.EDITOR] = " and ".join(record_dict[Fields.EDITOR])

            record_dict.pop("TY", None)
            record_dict.pop("Y2", None)
            record_dict.pop("DB", None)
            record_dict.pop("C1", None)
            record_dict.pop("T3", None)
            record_dict.pop("ER", None)

            for key, value in record_dict.items():
                record_dict[key] = str(value)

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            id_labeler=id_labeler,
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=self.review_manager.logger,
        )

        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".ris":
            return self._load_ris()

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for JSTOR"""

        return record



#!/usr/bin/env python3
"""Scripts for the load process."""
from __future__ import annotations

import typing
from pathlib import Path

import inquirer

import colrev.record.record
from colrev.constants import Colors
from colrev.constants import EndpointType

# pylint: disable=too-few-public-methods


class CLISourceAdder:
    """CLI utility to add SearchSources"""

    def __init__(self, *, search_operation: colrev.ops.search.Search) -> None:
        self.search_operation = search_operation
        self.review_manager = search_operation.review_manager
        self.package_manager = self.review_manager.get_package_manager()

    def _select_source(
        self,
        *,
        source_candidates: list,
    ) -> dict:
        choices = [
            (
                f"{heuristic_source['source_candidate'].endpoint} "
                f"(confidence: {round(heuristic_source['confidence'], 2)}): ",
                heuristic_source,
            )
            for heuristic_source in source_candidates
        ]
        questions = [
            inquirer.List(
                "source",
                message=f"{Colors.ORANGE}Select search source{Colors.END}:",
                choices=choices,
            ),
        ]
        selected_heuristic_source = inquirer.prompt(questions)["source"]
        return selected_heuristic_source

    def _select_source_from_heuristics(
        self, *, filename: Path, source_candidates: list
    ) -> colrev.settings.SearchSource:
        if 1 == len(source_candidates):
            heuristic_source_dict = source_candidates[0]
            self.search_operation.review_manager.logger.info(
                f"Automatically selected {heuristic_source_dict['source_candidate'].endpoint}"
            )
        else:
            heuristic_source_dict = self._select_source(
                source_candidates=source_candidates
            )
            self.search_operation.review_manager.logger.info(
                f"Selected {heuristic_source_dict['source_candidate'].endpoint}"
            )

        search_source_class = self.package_manager.get_package_endpoint_class(
            package_type=EndpointType.search_source,
            package_identifier=heuristic_source_dict["source_candidate"].endpoint,
        )
        endpoint = search_source_class(
            source_operation=self,
            settings=heuristic_source_dict["source_candidate"].get_dict(),
        )

        params = f"search_file={filename}"
        source = endpoint.add_endpoint(
            operation=self.search_operation,
            params=params,
        )
        return source

    def add_new_sources(self) -> typing.List[colrev.settings.SearchSource]:
        """Select the new source from the heuristic_result_list."""

        new_search_files = self.search_operation.get_new_search_files()
        sources_added = []

        for filename in new_search_files:
            self.search_operation.review_manager.logger.info(
                f"Discover and add new DB source: {filename}"
            )
            heuristic_list = self.search_operation.get_new_source_heuristic(filename)
            for source_candidates in heuristic_list:
                source = self._select_source_from_heuristics(
                    filename=filename, source_candidates=source_candidates
                )
                sources_added.append(source)
        return sources_added



#! /usr/bin/env python
"""Package interfaces."""
from __future__ import annotations

import typing
from pathlib import Path

import zope.interface

from colrev.constants import EndpointType

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.process.operation
    import colrev.record.record
    import colrev.settings
    from colrev.constants import SearchSourceHeuristicStatus

# pylint: disable=too-many-ancestors


# pylint: disable=too-few-public-methods
class GeneralInterface(zope.interface.Interface):  # pylint: disable=inherit-non-class
    """The General Interface for all package endpoints

    Each package endpoint must implement the following attributes (methods)"""

    ci_supported = zope.interface.Attribute(
        """Flag indicating whether the package can be run in
        continuous integration environments (e.g. GitHub Actions)"""
    )


# pylint: disable=too-few-public-methods
class ReviewTypeInterface(
    GeneralInterface, zope.interface.Interface
):  # pylint: disable=inherit-non-class
    """The PackageEndpoint interface for ReviewTypes"""

    # pylint: disable=no-self-argument
    def initialize(settings: dict) -> dict:  # type: ignore
        """Initialize the review type"""
        return settings  # pragma: no cover


class SearchSourceInterface(
    GeneralInterface, zope.interface.Interface
):  # pylint: disable=inherit-non-class
    """The PackageEndpoint interface for SearchSources"""

    settings_class = zope.interface.Attribute("""Class for the package settings""")
    source_identifier = zope.interface.Attribute(
        """Source identifier for search and provenance
        Retrieved records are identified through the source_identifier
        when they are added to/updated in the SearchAPIFeed"""
    )
    search_types = zope.interface.Attribute(
        """SearchTypes associated with the SearchSource"""
    )

    heuristic_status: SearchSourceHeuristicStatus = zope.interface.Attribute(
        """The status of the SearchSource heuristic"""
    )

    # pylint: disable=no-self-argument
    def heuristic(filename: Path, data: str):  # type: ignore
        """Heuristic to identify to which SearchSource a search file belongs (for DB searches)"""

    # pylint: disable=no-self-argument
    def add_endpoint(  # type: ignore
        operation: colrev.process.operation.Operation,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add the SearchSource as an endpoint based on a query (passed to colrev search -a)
        params:
        - search_file="..." to add a DB search
        """

    # pylint: disable=no-self-argument
    def search(rerun: bool) -> None:  # type: ignore
        """Run a search of the SearchSource"""

    # pylint: disable=no-self-argument
    def prep_link_md(  # type: ignore
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ):
        """Retrieve masterdata from the SearchSource"""

    # pylint: disable=no-self-argument
    def load(  # type: ignore
        load_operation: colrev.ops.load.Load,
    ) -> dict:
        """Load records from the SearchSource (and convert to .bib)"""

    # pylint: disable=no-self-argument
    def prepare(record: dict, source: colrev.settings.SearchSource) -> None:  # type: ignore
        """Run the custom source-prep operation"""


# pylint: disable=too-few-public-methods
class PrepInterface(
    GeneralInterface, zope.interface.Interface
):  # pylint: disable=inherit-non-class
    """The PackageEndpoint interface for prep operations"""

    settings_class = zope.interface.Attribute("""Class for the package settings""")
    source_correction_hint = zope.interface.Attribute(
        """Hint on how to correct metadata at source"""
    )

    always_apply_changes = zope.interface.Attribute(
        """Flag indicating whether changes should always be applied
        (even if the colrev_status does not transition to md_prepared)"""
    )

    # pylint: disable=no-self-argument
    def prepare(prep_record: dict) -> dict:  # type: ignore
        """Run the prep operation"""


# pylint: disable=too-few-public-methods
class PrepManInterface(
    GeneralInterface, zope.interface.Interface
):  # pylint: disable=inherit-non-class
    """The PackageEndpoint interface for prep-man operations"""

    settings_class = zope.interface.Attribute("""Class for the package settings""")

    # pylint: disable=no-self-argument
    def prepare_manual(records: dict) -> dict:  # type: ignore
        """Run the prep-man operation"""


# pylint: disable=too-few-public-methods
class DedupeInterface(
    GeneralInterface, zope.interface.Interface
):  # pylint: disable=inherit-non-class
    """The PackageEndpoint interface for dedupe operations"""

    settings_class = zope.interface.Attribute("""Class for the package settings""")

    # pylint: disable=no-self-argument
    # pylint: disable=no-method-argument
    def run_dedupe() -> None:  # type: ignore
        """Run the dedupe operation"""


# pylint: disable=too-few-public-methods
class PrescreenInterface(
    GeneralInterface, zope.interface.Interface
):  # pylint: disable=inherit-non-class
    """The PackageEndpoint interface for prescreen operations"""

    settings_class = zope.interface.Attribute("""Class for the package settings""")

    # pylint: disable=no-self-argument
    def run_prescreen(records: dict, split: list) -> dict:  # type: ignore
        """Run the prescreen operation"""


# pylint: disable=too-few-public-methods
class PDFGetInterface(
    GeneralInterface, zope.interface.Interface
):  # pylint: disable=inherit-non-class
    """The PackageEndpoint interface for pdf-get operations"""

    settings_class = zope.interface.Attribute("""Class for the package settings""")

    # pylint: disable=no-self-argument
    def get_pdf(record: dict) -> dict:  # type: ignore
        """Run the pdf-get operation"""
        return record  # pragma: no cover


# pylint: disable=too-few-public-methods
class PDFGetManInterface(
    GeneralInterface, zope.interface.Interface
):  # pylint: disable=inherit-non-class
    """The PackageEndpoint interface for pdf-get-man operations"""

    settings_class = zope.interface.Attribute("""Class for the package settings""")

    # pylint: disable=no-self-argument
    def pdf_get_man(records: dict) -> dict:
        """Run the pdf-get-man operation"""
        return records  # pragma: no cover


# pylint: disable=too-few-public-methods
class PDFPrepInterface(
    GeneralInterface, zope.interface.Interface
):  # pylint: disable=inherit-non-class
    """The PackageEndpoint interface for pdf-prep operations"""

    settings_class = zope.interface.Attribute("""Class for the package settings""")

    # pylint: disable=unused-argument
    # pylint: disable=no-self-argument
    def prep_pdf(  # type: ignore
        record: colrev.record.record_pdf.PDFRecord,
        pad: int,
    ) -> dict:
        """Run the prep-pdf operation"""
        return record.data  # pragma: no cover


# pylint: disable=too-few-public-methods
class PDFPrepManInterface(
    GeneralInterface, zope.interface.Interface
):  # pylint: disable=inherit-non-class
    """The PackageEndpoint interface for pdf-prep-man operations"""

    settings_class = zope.interface.Attribute("""Class for the package settings""")

    # pylint: disable=no-self-argument
    def pdf_prep_man(records: dict) -> dict:
        """Run the prep-man operation"""
        return records  # pragma: no cover


# pylint: disable=too-few-public-methods
class ScreenInterface(
    GeneralInterface, zope.interface.Interface
):  # pylint: disable=inherit-non-class
    """The PackageEndpoint interface for screen operations"""

    settings_class = zope.interface.Attribute("""Class for the package settings""")

    # pylint: disable=no-self-argument
    def run_screen(records: dict, split: list) -> dict:  # type: ignore
        """Run the screen operation"""


class DataInterface(
    GeneralInterface, zope.interface.Interface
):  # pylint: disable=inherit-non-class
    """The PackageEndpoint interface for data operations"""

    settings_class = zope.interface.Attribute("""Class for the package settings""")

    # pylint: disable=no-self-argument

    def update_data(  # type: ignore
        records: dict,
        synthesized_record_status_matrix: dict,
        silent_mode: bool,
    ) -> None:
        """
        Update the data by running the data operation. This includes data extraction,
        analysis, and synthesis.

        Parameters:
        records (dict): The records to be updated.
        synthesized_record_status_matrix (dict): The status matrix for the synthesized records.
        silent_mode (bool): Whether the operation is run in silent mode
        (for checks of review_manager/status).
        """

    def update_record_status_matrix(  # type: ignore
        synthesized_record_status_matrix: dict,
        endpoint_identifier: str,
    ) -> None:
        """Update the record status matrix,
        i.e., indicate whether the record is rev_synthesized for the given endpoint_identifier
        """

    # pylint: disable=no-method-argument
    def get_advice() -> dict:  # type: ignore
        """Get advice on how to operate the data package endpoint"""


ENDPOINT_OVERVIEW = {
    EndpointType.review_type: {
        "import_name": ReviewTypeInterface,
        "custom_class": "CustomReviewType",
        "operation_name": "operation",
    },
    EndpointType.search_source: {
        "import_name": SearchSourceInterface,
        "custom_class": "CustomSearchSource",
        "operation_name": "source_operation",
    },
    EndpointType.prep: {
        "import_name": PrepInterface,
        "custom_class": "CustomPrep",
        "operation_name": "prep_operation",
    },
    EndpointType.prep_man: {
        "import_name": PrepManInterface,
        "custom_class": "CustomPrepMan",
        "operation_name": "prep_man_operation",
    },
    EndpointType.dedupe: {
        "import_name": DedupeInterface,
        "custom_class": "CustomDedupe",
        "operation_name": "dedupe_operation",
    },
    EndpointType.prescreen: {
        "import_name": PrescreenInterface,
        "custom_class": "CustomPrescreen",
        "operation_name": "prescreen_operation",
    },
    EndpointType.pdf_get: {
        "import_name": PDFGetInterface,
        "custom_class": "CustomPDFGet",
        "operation_name": "pdf_get_operation",
    },
    EndpointType.pdf_get_man: {
        "import_name": PDFGetManInterface,
        "custom_class": "CustomPDFGetMan",
        "operation_name": "pdf_get_man_operation",
    },
    EndpointType.pdf_prep: {
        "import_name": PDFPrepInterface,
        "custom_class": "CustomPDFPrep",
        "operation_name": "pdf_prep_operation",
    },
    EndpointType.pdf_prep_man: {
        "import_name": PDFPrepManInterface,
        "custom_class": "CustomPDFPrepMan",
        "operation_name": "pdf_prep_man_operation",
    },
    EndpointType.screen: {
        "import_name": ScreenInterface,
        "custom_class": "CustomScreen",
        "operation_name": "screen_operation",
    },
    EndpointType.data: {
        "import_name": DataInterface,
        "custom_class": "CustomData",
        "operation_name": "data_operation",
    },
}

INTERFACE_MAP = {}
for endpoint_type, endpoint_data in ENDPOINT_OVERVIEW.items():
    import_name = endpoint_data["import_name"].__name__  # type: ignore
    INTERFACE_MAP[endpoint_type.name] = import_name



#! /usr/bin/env python
"""Creation of TEI as a PDF preparation operation"""
from __future__ import annotations

from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.env.utils
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PDFPrepInterface)
class GROBIDTEI:
    """Prepare PDFs by creating an annotated TEI document"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    TEI_PATH_RELATIVE = Path("data/.tei/")
    ci_supported: bool = Field(default=False)

    def __init__(
        self, *, pdf_prep_operation: colrev.ops.pdf_prep.PDFPrep, settings: dict
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.review_manager = pdf_prep_operation.review_manager

        if not pdf_prep_operation.review_manager.in_ci_environment():
            grobid_service = pdf_prep_operation.review_manager.get_grobid_service()
            grobid_service.start()
            self.tei_path = (
                pdf_prep_operation.review_manager.path / self.TEI_PATH_RELATIVE
            )
            self.tei_path.mkdir(exist_ok=True, parents=True)
            pdf_prep_operation.docker_images_to_stop.append(grobid_service.GROBID_IMAGE)

    def prep_pdf(
        self,
        record: colrev.record.record_pdf.PDFRecord,
        pad: int,  # pylint: disable=unused-argument
    ) -> dict:
        """Prepare the analysis of PDFs by creating a TEI (based on GROBID)"""

        if not record.data.get(Fields.FILE, "NA").endswith(".pdf"):
            return record.data

        if not record.get_tei_filename().is_file():
            self.review_manager.logger.debug(f" creating tei: {record.data['ID']}")
            _ = self.review_manager.get_tei(
                pdf_path=Path(record.data[Fields.FILE]),
                tei_path=record.get_tei_filename(),
            )

        return record.data



#! /usr/bin/env python
"""Cover-page removal as a PDF preparation operation"""
from __future__ import annotations

import shutil
import typing
from pathlib import Path

import pymupdf
import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
from colrev.constants import Fields
from colrev.constants import Filepaths

# pylint: disable=duplicate-code


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PDFPrepInterface)
class PDFCoverPage:
    """Prepare PDFs by removing unnecessary cover pages (e.g. researchgate, publishers)"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=False)

    def __init__(
        self,
        *,
        pdf_prep_operation: colrev.ops.pdf_prep.PDFPrep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.review_manager = pdf_prep_operation.review_manager

    def _check_scholarworks_first_page(self, *, page0: str, coverpages: list) -> None:
        if "followthisandadditionalworksat:https://scholarworks" in page0:
            coverpages.append(0)

    def _check_researchgate_first_page(self, *, page0: str, coverpages: list) -> None:
        if (
            "discussions,stats,andauthorprofilesforthispublicationat:"
            + "https://www.researchgate.net/publication"
            in page0
            or "discussions,stats,andauthorproﬁlesforthispublicationat:"
            + "https://www.researchgate.net/publication"
            in page0
        ):
            coverpages.append(0)

    def _check_jstor_first_page(self, *, page0: str, coverpages: list) -> None:
        if (
            "pleasecontactsupport@jstor.org.youruseofthejstorarchiveindicatesy"
            + "ouracceptanceoftheterms&conditionsofuse"
            in page0
            or "formoreinformationregardingjstor,pleasecontactsupport@jstor.org"
            in page0
        ):
            coverpages.append(0)

    def _check_emerald_first_page(self, *, page0: str, coverpages: list) -> None:
        if (
            "emeraldisbothcounter4andtransfercompliant.theorganizationisapartnero"
            "fthecommitteeonpublicationethics(cope)andalsoworkswithporticoandthe"
            "lockssinitiativefordigitalarchivepreservation.*relatedcontentand"
            "downloadinformationcorrectattimeofdownload" in page0
        ):
            coverpages.append(0)

    def _check_informs_first_page(
        self, *, page0: str, page1: str, coverpages: list
    ) -> None:
        if (
            "thisarticlewasdownloadedby" in page0
            and "fulltermsandconditionsofuse:" in page0
        ):
            coverpages.append(0)
        if (
            "thisarticlemaybeusedonlyforthepurposesofresearch" in page0
            and "abstract" not in page0
            and "keywords" not in page0
            and "abstract" in page1
            and "keywords" in page1
        ):
            coverpages.append(0)

    def _check_ais_first_page(self, *, page0: str, coverpages: list) -> None:
        if (
            "associationforinformationsystemsaiselectroniclibrary(aisel)" in page0
            and "abstract" not in page0
            and "keywords" not in page0
        ):
            coverpages.append(0)

    def _check_tandf_first_page(
        self, *, page0: str, page1: str, coverpages: list
    ) -> None:
        if ("pleasescrolldownforarticle" in page0) or ("viewrelatedarticles" in page0):
            if "abstract" not in page0 and "keywords" not in page0:
                coverpages.append(0)
                if (
                    "terms-and-conditions" in page1
                    and "abstract" not in page1
                    and "keywords" not in page1
                ):
                    coverpages.append(1)

    def _get_coverpages(
        self, record: colrev.record.record_pdf.PDFRecord
    ) -> typing.List[int]:
        coverpages: typing.List[int] = []

        doc = pymupdf.Document(str(record.data[Fields.FILE]))

        if doc.page_count == 1:
            return coverpages

        first_page_average_hash_16 = record.get_pdf_hash(page_nr=1, hash_size=16)

        # Note : to generate hashes from a directory containing single-page PDFs:
        # colrev pdf-prep --get_hashes path
        first_page_hashes = [
            "ffff83ff81ff81ffc3ff803f81ff80ffc03fffffffffffffffff96ff9fffffff",
            "ffffffffc0ffc781c007c007cfffc7ffffffffffffffc03fc007c003c827ffff",
            "84ff847ffeff83ff800783ff801f800180038fffffffffffbfff8007bffff83f",
            "83ff03ff03ff83ffffff807f807f9fff87ffffffffffffffffff809fbfffffff",
            "ffffffffc0ffc781c03fc01fffffffffffffffffffffc03fc01fc003ffffffff",
            "ffffffffc0ffc781c007c007cfffc7ffffffffffffffc03fc007c003ce27ffff",
            "ffffe7ffe3ffc3fffeff802780ff8001c01fffffffffffffffff80079ffffe3f",
            "ffffffffc0ffc781c00fc00fc0ffc7ffffffffffffffc03fc007c003cf27ffff",
            "ffffffffc0ffc781c007c003cfffcfffffffffffffffc03fc003c003cc27ffff",
            "ffffe7ffe3ffc3ffffff80e780ff80018001ffffffffffffffff93ff83ff9fff",
        ]

        if str(first_page_average_hash_16) in first_page_hashes:
            coverpages.append(0)

        res = doc.load_page(0).get_text()
        page0 = res.replace(" ", "").replace("\n", "").lower()

        res = doc.load_page(1).get_text()
        page1 = res.replace(" ", "").replace("\n", "").lower()

        # input(page0)

        self._check_scholarworks_first_page(page0=page0, coverpages=coverpages)
        self._check_researchgate_first_page(page0=page0, coverpages=coverpages)
        self._check_jstor_first_page(page0=page0, coverpages=coverpages)
        self._check_emerald_first_page(page0=page0, coverpages=coverpages)
        self._check_informs_first_page(page0=page0, page1=page1, coverpages=coverpages)
        self._check_ais_first_page(page0=page0, coverpages=coverpages)
        self._check_tandf_first_page(page0=page0, page1=page1, coverpages=coverpages)

        return list(set(coverpages))

    def prep_pdf(
        self,
        record: colrev.record.record_pdf.PDFRecord,
        pad: int,  # pylint: disable=unused-argument
    ) -> dict:
        """Prepare the PDF by removing coverpages (if any)"""

        if not record.data[Fields.FILE].endswith(".pdf"):
            return record.data

        cp_path = Filepaths.LOCAL_ENVIRONMENT_DIR / Path(".coverpages")
        cp_path.mkdir(exist_ok=True)

        coverpages = self._get_coverpages(record)
        if not coverpages:
            return record.data
        if coverpages:
            original = self.review_manager.path / Path(record.data[Fields.FILE])
            file_copy = self.review_manager.path / Path(
                record.data[Fields.FILE].replace(".pdf", "_with_cp.pdf")
            )
            shutil.copy(original, file_copy)
            record.extract_pages(
                pages=coverpages,
                save_to_path=cp_path,
            )
            self.review_manager.report_logger.info(
                f"removed cover page for ({record.data[Fields.ID]})"
            )
        return record.data



#! /usr/bin/env python
"""Convenience functions to load RIS files"""
from __future__ import annotations

import logging
import re
import typing
from pathlib import Path

import colrev.loader.loader


# pylint: disable=too-few-public-methods
# pylint: disable=too-many-arguments
# pylint: disable=too-many-instance-attributes


class NextLine(Exception):
    """NextLineException"""


class ParseError(Exception):
    """Parsing error"""


class RISLoader(colrev.loader.loader.Loader):
    """Loads ris files"""

    PATTERN = r"^[A-Z0-9]{2,4} "

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        *,
        filename: Path,
        entrytype_setter: typing.Callable,
        field_mapper: typing.Callable,
        id_labeler: typing.Callable,
        unique_id_field: str = "",
        logger: logging.Logger = logging.getLogger(__name__),
    ):
        super().__init__(
            filename=filename,
            id_labeler=id_labeler,
            unique_id_field=unique_id_field,
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=logger,
        )

        self.current: dict = {}
        self.pattern = re.compile(self.PATTERN)

    @classmethod
    def get_nr_records(cls, filename: Path) -> int:
        """Get the number of records in the file"""
        count = 0
        with open(filename, encoding="utf-8") as file:
            for line in file:
                if line.startswith("TY "):
                    count += 1
        return count

    def _get_tag(self, line: str) -> str:
        """Get the tag from a line in the RIS file."""
        return line[0 : line.find(" ")].rstrip()

    def _get_content(self, line: str) -> str:
        """Get the content from a line"""
        return line[line.find(" - ") + 3 :].strip()

    def _add_tag(self, tag: str, line: str) -> None:
        new_value = self._get_content(line)

        if tag not in self.current:
            self.current[tag] = new_value
        elif isinstance(self.current[tag], str):
            self.current[tag] = [self.current[tag], new_value]
        elif isinstance(self.current[tag], list):
            self.current[tag].append(new_value)

    def _parse_tag(self, line: str) -> dict:
        tag = self._get_tag(line)

        if tag.strip() == "":
            return self.current

        self._add_tag(tag, line)
        raise NextLine

    def _parse_lines(self, lines: list) -> typing.Iterator[dict]:
        for line in lines:
            try:
                yield self._parse_tag(line)
                self.current = {}
            except NextLine:
                continue

    def _clean_text(self, text: str) -> str:
        # Example:
        # Provider: JSTOR http://www.jstor.org
        # Database: JSTOR
        # Content: text/plain; charset="UTF-8"

        lines = []
        for line in text.split("\n"):
            if re.match(self.pattern, line):
                lines.append(line)
            if line.strip() in ["", "\n"]:
                lines.append(line)
        lines.append("")
        return "\n".join(lines)

    def load_records_list(self, *, content: str = "") -> list:
        """Load ris entries

        The resulting keys should coincide with those in the KEY_MAP
        but they can be adapted before calling the convert_to_records()"""

        # Note : depending on the source, a specific ris_parser implementation may be selected.
        # its DEFAULT_LIST_TAGS can be extended with list fields that should be joined automatically

        if content == "":
            content = self.filename.read_text(encoding="utf-8")
            content = self._clean_text(content)

        lines = content.split("\n")
        records_list = list(r for r in self._parse_lines(lines) if r)

        return records_list

    # def apply_ris_fixes(self) -> None:
    #     """Fix common defects in RIS files"""

    #     # Error to fix: for lists of keywords, each line should start with the KW tag

    #     with open(self.filename, encoding="UTF-8") as file:
    #         lines = [line.rstrip("\n") for line in file]
    #         # add missing start tags in lists (like KW)
    #         processing_tag = ""
    #         for i, line in enumerate(lines):
    #             tag_match = re.match(r"^[A-Z][A-Z0-9]+(\s+)-", line)  # |^ER\s?|^EF\s?
    #             if tag_match:
    #                 processing_tag = tag_match.group()
    #             elif line == "":
    #                 processing_tag = ""
    #                 continue
    #             elif processing_tag == "":
    #                 continue
    #             else:
    #                 lines[i] = f"{processing_tag} {line}"

    #     with open(self.filename, "w", encoding="utf-8") as file:
    #         for line in lines:
    #             file.write(f"{line}\n")



#! /usr/bin/env python
"""Consolidation of metadata based on the Pubmed API as a prep operation"""
from __future__ import annotations

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.packages.pubmed.src.pubmed as pubmed_connector
import colrev.record.record
from colrev.constants import Fields

# pylint: disable=duplicate-code


# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class PubmedMetadataPrep:
    """Prepares records based on Pubmed metadata"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=True)

    source_correction_hint = "ask the publisher to correct the metadata"
    always_apply_changes = False

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.prep_operation = prep_operation

        self.pubmed_source = pubmed_connector.PubMedSearchSource(
            source_operation=prep_operation
        )

        self.pubmed_prefixes = [
            s.get_origin_prefix()
            for s in prep_operation.review_manager.settings.sources
            if s.endpoint == "colrev.pubmed"
        ]

    def check_availability(
        self, *, source_operation: colrev.process.operation.Operation
    ) -> None:
        """Check status (availability) of the Pubmed API"""
        self.pubmed_source.check_availability(source_operation=source_operation)

    def prepare(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> colrev.record.record.Record:
        """Prepare a record based on Pubmed metadata"""

        if any(
            pubmed_prefix in o
            for pubmed_prefix in self.pubmed_prefixes
            for o in record.data[Fields.ORIGIN]
        ):
            # Already linked to a pubmed record
            return record

        self.pubmed_source.prep_link_md(
            prep_operation=self.prep_operation, record=record
        )
        return record



#! /usr/bin/env python
"""Consolidation of metadata based on Europe PMC API as a prep operation"""
from __future__ import annotations

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.packages.europe_pmc.src.europe_pmc as europe_pmc_connector
import colrev.record.record


# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class EuropePMCMetadataPrep:
    """Prepares records based on Europe PCM metadata"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=True)

    source_correction_hint = "ask the publisher to correct the metadata"
    always_apply_changes = False

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.prep_operation = prep_operation
        self.review_manager = prep_operation.review_manager
        self.epmc_source = europe_pmc_connector.EuropePMCSearchSource(
            source_operation=self.prep_operation
        )

    def prepare(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> colrev.record.record.Record:
        """Prepare a record based on Europe PMC metadata"""

        self.epmc_source.prep_link_md(prep_operation=self.prep_operation, record=record)
        return record
